from datetime import datetime
from enum import Enum
import logging
import os

from click import Choice, Option, Parameter, Path

# System information
CWD = os.getcwd()
if (NUM_CPUS := os.cpu_count()) is None:
    logging.warning("Failed to determine CPU count: defaulting to 1")
    NUM_CPUS = 1

DEFAULT_PHRED_ENC = 33
DEFAULT_MIN_PHRED = 25


class MateOrientationOption(Enum):
    """ Options of mate orientation for alignment with Bowtie2.

    See Bowtie2 manual for full documentation:
    https://bowtie-bio.sourceforge.net/bowtie2/manual.shtml
    """
    FR = "fr"
    RF = "rf"
    FF = "ff"


class AdapterSequence(Enum):
    """ Adapter sequences """
    ILLUMINA_3P = "AGATCGGAAGAGC"


# Input/output options
opt_out_dir = Option(("--out-dir",),
                     type=Path(file_okay=False),
                     default=os.path.join(".", "output"),
                     help="Where to output all finished files")
opt_temp_dir = Option(("--temp-dir",),
                      type=Path(file_okay=False),
                      default=os.path.join(".", "temp"),
                      help="Where to write all temporary files")
opt_save_temp = Option(("--save-temp/--erase-temp",),
                       type=bool,
                       default=False,
                       help=("Whether to save or erase temporary files "
                             "after the program exits"))

# Resource usage options
opt_parallel = Option(("--parallel/--no-parallel",),
                      type=bool,
                      default=True,
                      help="Whether to run multiple jobs in parallel")
opt_max_procs = Option(("--max-procs",),
                       type=int,
                       default=NUM_CPUS,
                       help="Maximum number of simultaneous processes")

# Experiment and analysis setup options
opt_library = Option(("--library",),
                     type=Path(dir_okay=False),
                     default="",
                     help="Library CSV file")
opt_samples = Option(("--samples",),
                     type=Path(dir_okay=False),
                     default="",
                     help="Samples file")
opt_rerun = Option(("--rerun/--no-rerun",),
                   type=bool,
                   default=False,
                   help="Whether to regenerate files that already exist")

# Reference sequence (FASTA) files
opt_fasta = Option(("--fasta",),
                   type=Path(exists=True, dir_okay=False),
                   help="FASTA file of all reference sequences in the project")

# Sequencing read (FASTQ) files
opt_fastqs = Option(("--fastqs",),
                    type=Path(exists=True, dir_okay=False),
                    multiple=True,
                    default=(),
                    help="FASTQ files of single-end reads")
opt_fastqi = Option(("--fastqi",),
                    type=Path(exists=True, dir_okay=False),
                    multiple=True,
                    default=(),
                    help="FASTQ files of interleaved paired reads")
opt_fastq1 = Option(("--fastq1",),
                    type=Path(exists=True, dir_okay=False),
                    multiple=True,
                    default=(),
                    help="FASTQ files of mate 1 paired-end reads")
opt_fastq2 = Option(("--fastq2",),
                    type=Path(exists=True, dir_okay=False),
                    multiple=True,
                    default=(),
                    help="FASTQ files of mate 2 paired-end reads")

# Sequencing read (FASTQ/BAM) options
opt_phred_enc = Option(("--phred-enc",),
                       type=int,
                       default=DEFAULT_PHRED_ENC,
                       help="Phred score encoding in FASTQ/SAM/BAM files")
opt_min_phred = Option(("--min-phred",),
                       type=int,
                       default=DEFAULT_MIN_PHRED,
                       help="Minimum Phred score to use a base call")
opt_fastqc = Option(("--fastqc/--no-fastqc",),
                    type=bool,
                    default=True,
                    help="Whether to check quality of FASTQ files")
opt_qc_extract = Option(("--qc-extract/--qc-no-extract",),
                        type=bool,
                        default=False,
                        help="Whether to unzip FASTQC reports")

# Demultiplexing options

opt_demultiplex = Option(("--demult-on/--demult-off",),
                         type=bool,
                         default=False,
                         help="Whether to run demultiplexing")

opt_parallel_demultiplexing = Option(("--parallel-demultiplexing",),
                                     type=bool,
                                     default=False,
                                     help="Whether to run demultiplexing at maximum speed by submitting multithreaded grep functions")

opt_clipped_demultiplexing = Option(("--clipped",),
                                    type=int,
                                    default=0,
                                    help="Designates the amount of clipped patterns to search for in the sample, will raise compution time")

opt_mismatch_tolerence = Option(("--mismatch-tolerence",),
                                type=int,
                                default=0,
                                help="Designates the allowable amount of mismatches allowed in a string and still be considered a valid pattern find. \
                            will increase non-parallel computation at a factorial rate. use caution going above 2 mismatches. does not apply to clipped sequences.")

opt_index_tolerence = Option(("--index-tolerance",),
                             type=int,
                             default=0,
                             help="Designates the allowable amount of distance you allow the pattern to be found in a read from the reference index")

opt_barcode_start = Option(("--barcode-start",),
                           type=int,
                           default=0,
                           help="index of start of barcode")
opt_barcode_length = Option(("--barcode-length",),
                            type=int,
                            default=0,
                            help="length of barcode")
opt_demulti_overwrite = Option(("--demulti-overwrite",),
                               type=bool,
                               default=False,
                               help="desiginates whether to overwrite the grepped fastq. should only be used if changing setting on the same sample")

# Demultiplexed sequencing read (FASTQ) directories
opt_fastqs_dir = Option(("--fastqs-dir",),
                        type=Path(exists=True, file_okay=False),
                        multiple=True,
                        default=(),
                        help="Directory containing demultiplexed FASTQ files of single-end reads from one sample")
opt_fastqi_dir = Option(("--fastqi-dir",),
                        type=Path(exists=True, file_okay=False),
                        multiple=True,
                        default=(),
                        help="Directory containing demultiplexed FASTQ files of interleaved paired-end reads from one sample")
opt_fastq12_dir = Option(("--fastq12-dir",),
                         type=Path(exists=True, file_okay=False),
                         multiple=True,
                         default=(),
                         help="Directory containing demultiplexed pairs of FASTQ files of mate 1 and mate 2 reads from one sample")

# Alignment map (BAM) files
opt_bamf = Option(("--bamf",),
                  type=Path(exists=True, dir_okay=False),
                  multiple=True,
                  default=(),
                  help="BAM file")
opt_bamd = Option(("--bamd",),
                  type=Path(exists=True, file_okay=False),
                  multiple=True,
                  default=(),
                  help="Directory of BAM files")

# Adapter trimming options with Cutadapt
opt_cutadapt = Option(("--cut/--no-cut",),
                      type=bool,
                      default=True,
                      help="Whether to trim reads with Cutadapt before alignment")
opt_cut_q1 = Option(("--cut-q1",),
                    type=int,
                    default=DEFAULT_MIN_PHRED,
                    help="Phred score for read 1 quality trimming")
opt_cut_q2 = Option(("--cut-q2",),
                    type=int,
                    default=DEFAULT_MIN_PHRED,
                    help="Phred score for read 2 quality trimming")
opt_cut_g1 = Option(("--cut-g1",),
                    type=str,
                    multiple=True,
                    default=(),
                    help="5' adapter for read 1")
opt_cut_a1 = Option(("--cut-a1",),
                    type=str,
                    multiple=True,
                    default=(AdapterSequence.ILLUMINA_3P.value,),
                    help="3' adapter for read 1")
opt_cut_g2 = Option(("--cut-g2",),
                    type=str,
                    multiple=True,
                    default=(),
                    help="5' adapter for read 2")
opt_cut_a2 = Option(("--cut-a2",),
                    type=str,
                    multiple=True,
                    default=(AdapterSequence.ILLUMINA_3P.value,),
                    help="3' adapter for read 2")
opt_cut_o = Option(("--cut-O",),
                   type=int,
                   default=6,
                   help="Minimum overlap of read and adapter")
opt_cut_e = Option(("--cut-e",),
                   type=float,
                   default=0.1,
                   help="Error tolerance for adapters")
opt_cut_indels = Option(("--cut-indels/--cut-no-indels",),
                        type=bool,
                        default=True,
                        help="Whether to allow indels in adapters")
opt_cut_nextseq = Option(("--cut-nextseq/--cut-no-nextseq",),
                         type=bool,
                         default=False,
                         help="Whether to trim high-quality Gs from 3' end")
opt_cut_discard_trimmed = Option(
    ("--cut-discard-trimmed/--cut-keep-trimmed",),
    type=bool,
    default=False,
    help="Whether to discard reads in which an adapter was found")
opt_cut_discard_untrimmed = Option(
    ("--cut-discard-untrimmed/--cut-keep-untrimmed",),
    type=bool,
    default=False,
    help="Whether to discard reads in which no adapter was found")
opt_cut_m = Option(("--cut-m",),
                   type=int,
                   default=20,
                   help="Discard reads shorter than this length after trimming")

# Alignment options with Bowtie2
opt_bt2_local = Option(("--bt2-local/--bt2-end-to-end",),
                       type=bool,
                       default=True,
                       help="Whether to perform local or end-to-end alignment")
opt_bt2_discordant = Option(("--bt2-discordant/--bt2-no-discordant",),
                            type=bool,
                            default=False,
                            help="Whether to output discordant alignments")
opt_bt2_mixed = Option(("--bt2-mixed/--bt2-no-mixed",),
                       type=bool,
                       default=False,
                       help="Whether to align individual mates of unaligned pairs")
opt_bt2_dovetail = Option(("--bt2-dovetail/--bt2-no-dovetail",),
                          type=bool,
                          default=False,
                          help="Whether to treat dovetailed mate pairs as concordant")
opt_bt2_contain = Option(("--bt2-contain/--bt2-no-contain",),
                         type=bool,
                         default=True,
                         help="Whether to treat nested mate pairs as concordant")
opt_bt2_unal = Option(("--bt2-unal/--bt2-no-unal",),
                      type=bool,
                      default=False,
                      help="Whether to output unaligned reads")
opt_bt2_i = Option(("--bt2-I",),
                   type=int,
                   default=0,
                   help="Minimum fragment length for valid paired-end alignments")
opt_bt2_x = Option(("--bt2-X",),
                   type=int,
                   default=600,
                   help="Maximum fragment length for valid paired-end alignments")
opt_bt2_score_min = Option(("--bt2-score-min",),
                           type=str,
                           default="L,1,0.5",
                           help="Minimum score for a valid alignment")
opt_bt2_s = Option(("--bt2-i", "bt2_s"),
                   type=str,
                   default="L,1,0.1",
                   help="Seed interval")
opt_bt2_l = Option(("--bt2-L",),
                   type=int,
                   default=12,
                   help="Seed length")
opt_bt2_gbar = Option(("--bt2-gbar",),
                      type=int,
                      default=4,
                      help="Minimum distance of a gap from end of a read")
opt_bt2_d = Option(("--bt2-D",),
                   type=int,
                   default=4,
                   help="Maximum number of failed seed extensions")
opt_bt2_r = Option(("--bt2-R",),
                   type=int,
                   default=2,
                   help="Maximum number of times to re-seed")
opt_bt2_dpad = Option(("--bt2-dpad",),
                      type=int,
                      default=2,
                      help="Width of padding on alignment matrix, to allow gaps")
opt_bt2_orient = Option(("--bt2-orient",),
                        type=Choice(tuple(op.value for op
                                          in MateOrientationOption),
                                    case_sensitive=False),
                        default=MateOrientationOption.FR.value,
                        help="Valid orientations of paired-end mates")

# Reference section specification options
opt_coords = Option(("--coords", "-c"),
                    type=(str, int, int),
                    multiple=True,
                    default=(),
                    help=("Reference name, 5' end, and 3' end of a section; "
                          "coordinates are 1-indexed and include both ends"))
opt_primers = Option(("--primers", "-p"),
                     type=(str, str, str),
                     multiple=True,
                     default=(),
                     help=("Reference name, forward primer, and reverse primer "
                           "of a section; reverse primer must be given 5' to 3'"))
opt_primer_gap = Option(("--primer-gap",),
                        type=int,
                        default=2,
                        help=("Number of bases to leave as a gap between the "
                              "end of a primer and the end of the section"))
opt_autosect = Option(("--autosect/--no-autosect",),
                      type=bool,
                      default=False,
                      help=("Whether, for every reference that was not explicitly "
                            "given at least one section (by --initial_coords or "
                            "--primers), to generate coordinates covering the "
                            "entire reference sequence automatically"))

# Vectoring options
opt_batch_size = Option(("--batch-size", "-z"),
                        type=float,
                        default=32.0,
                        help=("Maximum size of each batch of mutation vectors, "
                              "in millions of base calls"))
opt_strict_pairs = Option(("--strict-pairs/--no-strict-pairs",),
                          type=bool,
                          default=True,
                          help=("Whether to require that every paired "
                                "read that maps to a section also have "
                                "a mate that maps to the section"))
opt_ambid = Option(("--ambid/--no-ambid",),
                   type=bool,
                   default=True,
                   help=("Whether to find and label all ambiguous "
                         "insertions and deletions (improves accuracy "
                         "but runs slower)"))

# Mutational profile report files
opt_report = Option(("--mp-report",),
                    type=Path(exists=True, dir_okay=False),
                    multiple=True,
                    default=(),
                    help="Path to the bit vector folder or list of paths to the bit vector folders.")

# Clustering options
opt_cluster = Option(("--clust/--cluster-off",),
                     type=bool,
                     default=False,
                     help="Whether to run clustering")
opt_max_clusters = Option(("--max-clusters",),
                          type=int,
                          default=3,
                          help='Maximum number of clusters.')
opt_min_iter = Option(("--min-iter",),
                      type=int,
                      default=100,
                      help='Minimum number of iteration before checking convergence of EM.')
opt_max_iter = Option(("--max-iter",),
                      type=int,
                      default=500,
                      help='Maximum number of iteration before stopping EM.')
opt_signal_thresh = Option(("--signal-thresh",),
                           type=float,
                           default=0.005,
                           help='Minimum Mutation fraction to keep a base.')
opt_info_thresh = Option(("--info-thresh",),
                         type=float,
                         default=0.05)
opt_include_gu = Option(("--include-gu/--exclude-gu",),
                        type=bool,
                        default=False,
                        help='Whether to include G and U bases in reads.')
opt_polya_max = Option(("--polya-max",),
                       type=int,
                       default=4,
                       help='Maximum length of poly(A) sequences to include.')
opt_include_del = Option(("--include-del/--exclude-del",),
                         type=bool,
                         default=False,
                         help='Whether to include deletions in reads.')
opt_min_reads = Option(("--min-reads",),
                       type=int,
                       default=1000,
                       help='Minimum number of reads to start clustering.')
opt_convergence_cutoff = Option(("--convergence-cutoff",),
                                type=float,
                                default=0.5,
                                help='Minimum difference between the log-likelihood of two consecutive iterations to stop EM.')
opt_num_runs = Option(("--num-runs", "-n"),
                      type=int,
                      default=10,
                      help='Number of time to run the clustering algorithm.')

# Aggregation

opt_bv_files = Option(("--bv-files", "-bv"),
                      type=Path(exists=True, dir_okay=True, file_okay=False),
                      multiple=True,
                      default=(),
                      help="Tuple of paths. Give the path to the sample folder to process every section. Give the path to a report to process a single section.")

opt_clustering_file = Option(("--clustering-file", "-cf"),
                             # type=Path(exists=True, dir_okay=False),
                             default='',
                             help="Path to the json clustering file from dreem clustering.")

opt_rnastructure_path = Option(("--rnastructure-path", "-rs"),
                               type=Path(),
                               help='Path to the RNAstructure executable folder (e.g. /home/user/RNAstructure/exe/). Use this option if RNAstructure is not in your PATH.',
                               default='')
opt_rnastructure_use_temp = Option(("--rnastructure-use-temp", "-rst"),
                                   type=bool, default=False,
                                   help='Use the temperature signal to make predictions with RNAstructure')
opt_rnastructure_fold_args = Option(("--rnastructure-fold-args", "-rsa"),
                                    type=str,
                                    default='',
                                    help="Additional arguments to pass to RNAstructure's Fold command")

opt_rnastructure_use_dms = Option(("--rnastructure-use-dms", "-rsd"),
                                  type=bool,
                                  default=False,
                                  help="Use the DMS signal to make predictions with RNAstructure")
opt_rnastructure_dms_min_unpaired_value = Option(("--rnastructure-dms-min-unpaired-value", "-rsdmin"),
                                                 type=int,
                                                 default=0.07,
                                                 help="Minimum unpaired value for using the dms signal as an input for RNAstructure")
opt_rnastructure_dms_max_paired_value = Option(("--rnastructure-dms-max-paired-value", "-rsdmax"),
                                               type=int,
                                               default=0.01,
                                               help="Maximum paired value for using the dms signal as an input for RNAstructure")
opt_rnastructure_deltag_ensemble = Option(("--rnastructure-deltag-ensemble", "-rspa"),
                                          type=bool,
                                          default=False,
                                          help="Use RNAstructure partition function to predict free energy")
opt_rnastructure_probability = Option(("--rnastructure_probability", "-rspr"),
                                      type=bool,
                                      default=False,
                                      help="Use RNAstructure partition function to predict per-base pairing probability")

# Drawing options

opt_draw_input = Option(("--inpt",),
                        multiple=True,
                        type=Path(exists=True, dir_okay=False),
                        default=(),
                        help="Path to a dreem output format file. Can be specified multiple times.")

opt_section = Option(("--section",),
                        multiple=True,
                        default=('full',),
                        help="Section to draw. Can be specified multiple times.")

opt_flat = Option(("--flat/--no-flat",),
                  is_flag=True,
                  default=True,
                  help="Flatten the output folder structure. This names your files [reference]__[section]__[plot_name].html")

opt_mutation_fraction = Option(("--mutation_fraction", "-mf"),
                               is_flag=True,
                               default=True,
                               help="Plot mutation_fraction plot. See Plots/gallery.")

opt_mutation_fraction_identity = Option(("--mutation_fraction_identity", "-mfi"),
                                        is_flag=True,
                                        default=True,
                                        help="Plot mutation_fraction_identity plot. See Plots/gallery.")

opt_base_coverage = Option(("--base_coverage", "-bc"),
                           is_flag=True,
                           default=True,
                           help="Plot base_coverage plot. See Plots/gallery.")

opt_mutations_in_barcodes = Option(("--mutations_in_barcodes", "-mib"),
                                   is_flag=True,
                                   default=False,
                                   help="Plot mutations_in_barcodes plot. See Plots/gallery.")

opt_mutations_per_read_per_sample = Option(("--mutation_per_read_per_reference", "-mprps"),
                                           is_flag=True,
                                           default=True,
                                           help="Plot mutation_per_read_per_reference plot. See Plots/gallery.")

# Logging options
opt_verbose = Option(("--verbose", "-v"),
                     count=True,
                     help="Print info or info+debug messages to stdout")
opt_quiet = Option(("--quiet",),
                   count=True,
                   help="Suppress warnings or warnings+errors to stdout")
opt_log = Option(("--log",),
                 type=Path(exists=False, dir_okay=False),
                 default=os.path.join(CWD, datetime.now().strftime(
                     "dreem_%Y-%m-%d_%H:%M:%S.log")))
opt_profile = Option(("--profile",),
                     type=Path(exists=False, dir_okay=False),
                     default="",
                     help="Profile code performance and log results to file")

# Misc
opt_version = Option(("--version",),
                     is_flag=True,
                     help="Show the version and exit.")


def merge_params(*param_lists: list[Parameter]):
    """ Merge lists of Click parameters, dropping duplicates. """
    params = list()
    names = set()
    for param_list in param_lists:
        for param in param_list:
            if param.name not in names:
                params.append(param)
                names.add(param.name)
    return params



from functools import partial
from itertools import chain, starmap as itsmap
from logging import getLogger
from multiprocessing import Pool
from pathlib import Path
from typing import Iterable

from .fqs import FastqUnit, run_fastqc, run_cutadapt, run_bowtie2
from .xams import dedup_sam, sort_xam, view_xam, index_bam
from ..util import path
from ..util.parallel import get_num_parallel
from ..util.seq import parse_fasta, write_fasta
from ..util.shell import BOWTIE2_BUILD_CMD, run_cmd


logger = getLogger(__name__)


def get_fasta_index_paths(fasta: Path):
    """ Return the Bowtie 2 index paths for a FASTA file. """
    return [fasta.with_suffix(ext) for ext in path.BOWTIE2_INDEX_EXTS]


def index_fasta_file(fasta: Path,
                     prefix: Path,
                     n_procs: int = 1):
    """ Build a Bowtie2 index of a FASTA file. """
    logger.info(f"Began building Bowtie2 index of FASTA {fasta}")
    # Generate and run the command. Use quiet mode because otherwise,
    # Bowtie2-Build produces extremely verbose output.
    cmd = [BOWTIE2_BUILD_CMD, "-q", "--threads", n_procs, fasta, prefix]
    run_cmd(cmd)
    logger.info(f"Ended building Bowtie2 index of FASTA {fasta}: {prefix}")


def write_temp_ref_files(temp_dir: Path,
                         refset_path: Path,
                         refs: set[str],
                         n_procs: int):
    """ Write temporary FASTA files, each containing one reference that
    corresponds to a FASTQ file from demultiplexing. """
    ref_paths: dict[str, tuple[Path, Path]] = dict()
    if refs:
        # Parse the FASTA only if there are any references to write.
        for ref, seq in parse_fasta(refset_path):
            if ref in refs:
                # Write the reference sequence to a temporary FASTA file
                # only if at least one demultiplexed FASTQ file uses it.
                ref_path = path.OneRefSeqTempFilePath(top=str(temp_dir),
                                                      module=path.Module.ALIGN.value,
                                                      step=path.Step.ALIGN_REFS.value,
                                                      ref=ref,
                                                      ext=refset_path.suffix).path
                # Create the parent directory.
                ref_path.parent.mkdir(parents=True, exist_ok=True)
                logger.debug(f"Created directory: {ref_path.parent}")
                try:
                    # Write the temporary FASTA file.
                    write_fasta(ref_path, [(ref, seq)])
                    # Build a Bowtie2 index of the temporary FASTA file.
                    index_prefix = ref_path.with_suffix("")
                    index_fasta_file(ref_path, index_prefix, n_procs)
                except Exception as error:
                    logger.critical(
                        f"Failed to generate reference {ref_path}: {error}")
                else:
                    # Record the temporary FASTA and index prefix.
                    ref_paths[ref] = ref_path, index_prefix
    if missing := sorted(refs - set(ref_paths.keys())):
        logger.critical(f"Missing references in {refset_path}: "
                        + ", ".join(missing))
    return ref_paths


def fq_pipeline(fq_inp: FastqUnit,
                fasta: Path,
                bowtie2_index: Path, *,
                n_procs: int,
                out_dir: str,
                temp_dir: str,
                save_temp: bool,
                fastqc: bool,
                qc_extract: bool,
                cut: bool,
                cut_q1: int,
                cut_q2: int,
                cut_g1: str,
                cut_a1: str,
                cut_g2: str,
                cut_a2: str,
                cut_o: int,
                cut_e: float,
                cut_indels: bool,
                cut_nextseq: bool,
                cut_discard_trimmed: bool,
                cut_discard_untrimmed: bool,
                cut_m: int,
                bt2_local: bool,
                bt2_discordant: bool,
                bt2_mixed: bool,
                bt2_dovetail: bool,
                bt2_contain: bool,
                bt2_unal: bool,
                bt2_score_min: str,
                bt2_i: int,
                bt2_x: int,
                bt2_gbar: int,
                bt2_l: int,
                bt2_s: str,
                bt2_d: int,
                bt2_r: int,
                bt2_dpad: int,
                bt2_orient: str) -> list[Path]:
    """ Run all steps of the alignment pipeline for one FASTQ file or
    one pair of mated FASTQ files. """
    # Number of additional processes to use.
    n_padd = n_procs - 1
    # Get the name of the set of references.
    refset = path.RefsetSeqInFilePath.parse(fasta).refset
    # Run FASTQC on the input.
    if fastqc:
        fastqc_out = path.FastqcOutDirPath(top=str(out_dir),
                                           module=path.Module.ALIGN,
                                           sample=fq_inp.sample,
                                           fastqc=path.Fastqc.QC_INPUT).path
        run_fastqc(fq_inp, fastqc_out, qc_extract)
    # Trim adapters and low-quality bases.
    if cut:
        fq_cut = fq_inp.edit(top=str(temp_dir),
                             module=path.Module.ALIGN,
                             step=path.Step.ALIGN_TRIM)
        run_cutadapt(fq_inp, fq_cut,
                     n_procs=n_procs,
                     cut_q1=cut_q1,
                     cut_q2=cut_q2,
                     cut_g1=cut_g1,
                     cut_a1=cut_a1,
                     cut_g2=cut_g2,
                     cut_a2=cut_a2,
                     cut_o=cut_o,
                     cut_e=cut_e,
                     cut_indels=cut_indels,
                     cut_nextseq=cut_nextseq,
                     cut_discard_trimmed=cut_discard_trimmed,
                     cut_discard_untrimmed=cut_discard_untrimmed,
                     cut_m=cut_m)
        # Run FASTQC after trimming.
        if fastqc:
            fastqc_out = path.FastqcOutDirPath(top=str(out_dir),
                                               module=path.Module.ALIGN,
                                               sample=fq_inp.sample,
                                               fastqc=path.Fastqc.QC_TRIM).path
            run_fastqc(fq_cut, fastqc_out, qc_extract)
    else:
        fq_cut = None
    # Align to reference.
    sam_aligned = path.RefsetAlignmentTempFilePath(top=str(temp_dir),
                                                   module=path.Module.ALIGN,
                                                   step=path.Step.ALIGN_ALIGN,
                                                   sample=fq_inp.sample,
                                                   refset=refset,
                                                   ext=path.SAM_EXT).path
    run_bowtie2(fq_inp if fq_cut is None else fq_cut,
                bowtie2_index,
                sam_aligned,
                n_procs=n_procs,
                bt2_local=bt2_local,
                bt2_discordant=bt2_discordant,
                bt2_mixed=bt2_mixed,
                bt2_dovetail=bt2_dovetail,
                bt2_contain=bt2_contain,
                bt2_unal=bt2_unal,
                bt2_score_min=bt2_score_min,
                bt2_i=bt2_i,
                bt2_x=bt2_x,
                bt2_gbar=bt2_gbar,
                bt2_l=bt2_l,
                bt2_s=bt2_s,
                bt2_d=bt2_d,
                bt2_r=bt2_r,
                bt2_dpad=bt2_dpad,
                bt2_orient=bt2_orient)
    if not save_temp and fq_cut is not None:
        for fq_file in fq_cut.paths:
            fq_file.path.unlink(missing_ok=True)
    # Deduplicate.
    sam_deduped = path.RefsetAlignmentTempFilePath(top=str(temp_dir),
                                                   module=path.Module.ALIGN,
                                                   step=path.Step.ALIGN_DEDUP,
                                                   sample=fq_inp.sample,
                                                   refset=refset,
                                                   ext=path.SAM_EXT).path
    dedup_sam(sam_aligned, sam_deduped)
    if not save_temp:
        sam_aligned.unlink(missing_ok=True)
    # Sort by coordinate.
    bam_sorted = path.RefsetAlignmentTempFilePath(top=str(temp_dir),
                                                  module=path.Module.ALIGN,
                                                  step=path.Step.ALIGN_SORT,
                                                  sample=fq_inp.sample,
                                                  refset=refset,
                                                  ext=path.BAM_EXT).path
    sort_xam(sam_deduped, bam_sorted, name=False, n_padd=n_padd)
    if not save_temp:
        sam_deduped.unlink(missing_ok=True)
    # Split into one BAM file for each reference.
    index_bam(bam_sorted, n_padd=n_padd)  # Splitting requires an index.
    bams_out: list[Path] = list()
    for ref, _ in parse_fasta(fasta):
        bam_split = path.OneRefAlignmentOutFilePath(top=str(out_dir),
                                                    module=path.Module.ALIGN,
                                                    sample=fq_inp.sample,
                                                    ref=ref,
                                                    ext=path.BAM_EXT).path
        view_xam(bam_sorted, bam_split, n_padd=n_padd, ref=ref)
        bams_out.append(bam_split)
    if not save_temp:
        bam_sorted.unlink(missing_ok=True)
    # Return a list of the split BAM files.
    return bams_out


def fqs_pipeline(fq_units: list[FastqUnit],
                 main_fasta: Path, *,
                 max_procs: int,
                 parallel: bool,
                 out_dir: Path,
                 temp_dir: Path,
                 save_temp: bool,
                 **kwargs) -> list[Path]:
    """ Run all steps of alignment for one or more FASTQ files or pairs
    of mated FASTQ files. """
    # Validate the maximum number of processes.
    if max_procs < 1:
        logger.warning("Maximum CPUs must be ≥ 1: setting to 1")
        max_procs = 1
    # Get the name of the reference for every demultiplexed FASTQ.
    temp_refs = set(filter(None, (fq_unit.ref for fq_unit in fq_units)))
    # Write a temporary FASTA file and Bowtie2 index for each
    # demultiplexed FASTQ.
    temp_fasta_paths = write_temp_ref_files(temp_dir, main_fasta,
                                            temp_refs, max_procs)
    # Check if the main FASTA file already has a Bowtie2 index.
    if all(index.is_file() for index in get_fasta_index_paths(main_fasta)):
        # Bowtie2 index for the main FASTA exists.
        main_index = main_fasta.with_suffix("")
    else:
        # Bowtie2 index does not already exist.
        main_index = None
    try:
        # Make the arguments for each alignment task.
        iter_args: list[tuple[FastqUnit, Path, Path]] = list()
        # One alignment task will be created for each FASTQ unit.
        for fq in fq_units:
            if fq.ref:
                # If the FASTQ came from demultiplexing (so contains
                # reads from only one reference), then align to the
                # temporary FASTA file containing only that reference.
                try:
                    temp_fasta, temp_index = temp_fasta_paths[fq.ref]
                except KeyError:
                    # If the FASTA with that reference does not exist,
                    # then log an error and skip this FASTQ.
                    logger.critical(
                        f"Skipped FASTQ {fq} because reference '{fq.ref}' "
                        f"was not found in FASTA file {main_fasta}")
                    continue
                # Add these arguments to the lists of arguments that
                # will be passed to fq_pipeline.
                iter_args.append((fq, temp_fasta, temp_index))
                logger.debug(f"Added task: align {fq} to {temp_index}")
            else:
                # If the FASTQ may contain reads from ≥ 1 references,
                # then align to the FASTA file with all references.
                if main_index is None:
                    # The FASTA of all the references does not already
                    # have a Bowtie2 index, so build a temporary index.
                    # Determine the name of the set of references.
                    refset = path.RefsetSeqInFilePath.parse(main_fasta).refset
                    # Determine the path of the temporary Bowtie2 index
                    # of the main FASTA file.
                    main_index = path.RefsetBowtie2IndexTempPrefix(
                        top=str(temp_dir), module=path.Module.ALIGN.value,
                        step=path.Step.ALIGN_REFS.value, refset=refset).path
                    # Make its parent directory if it does not exist.
                    main_index.parent.mkdir(parents=True, exist_ok=True)
                    logger.debug(f"Created directory: {main_index.parent}")
                    # Build the Bowtie2 index.
                    index_fasta_file(main_fasta, main_index, max_procs)
                    # Create a symbolic link to the reference file in
                    # the same directory as the new index.
                    fasta_link = main_index.with_suffix(main_fasta.suffix)
                    fasta_link.symlink_to(main_fasta)
                    # Add the FASTA link and the Bowtie2 index to the
                    # set of files to delete after alignment finishes.
                    # Being deleted is the only purpose of fasta_link.
                    temp_fasta_paths[refset] = fasta_link, main_index
                # Add these arguments to the lists of arguments that
                # will be passed to fq_pipeline. Note that main_index
                # could be a pre-built index in the same directory as
                # main_fasta or a temporary index that is deleted when
                # alignment finishes; but only in the latter case is it
                # added to temp_fasta_paths.
                iter_args.append((fq, main_fasta, main_index))
                logger.debug(f"Added task: align {fq} to {main_index}")
        # Determine how to parallelize each alignment task.
        n_tasks_parallel, n_procs_per_task = get_num_parallel(len(fq_units),
                                                              max_procs,
                                                              parallel,
                                                              hybrid=True)
        # Pass the keyword arguments to every call of fq_pipeline.
        partial_fq_pipeline = partial(fq_pipeline,
                                      **{**dict(n_procs=n_procs_per_task,
                                                out_dir=out_dir,
                                                temp_dir=temp_dir,
                                                save_temp=save_temp),
                                         **kwargs})
        if n_tasks_parallel > 1:
            # Process the FASTQ files simultaneously.
            logger.debug(f"Initializing pool of {n_tasks_parallel} processes")
            with Pool(n_tasks_parallel) as pool:
                logger.debug(f"Opened pool of {n_tasks_parallel} processes")
                tasks = pool.starmap(partial_fq_pipeline, iter_args)
                bams = list(chain(*tasks))
            logger.debug(f"Closed pool of {n_tasks_parallel} processes")
        else:
            # Process the FASTQ files sequentially.
            tasks = itsmap(partial_fq_pipeline, iter_args)
            bams = list(chain(*tasks))
    finally:
        if not save_temp:
            # Delete the temporary files before exiting.
            for ref_file, index_prefix in temp_fasta_paths.values():
                # Reference file
                ref_file.unlink(missing_ok=True)
                logger.debug(f"Deleted temporary reference file: {ref_file}")
                # Index files
                for index_file in get_fasta_index_paths(index_prefix):
                    index_file.unlink(missing_ok=True)
                    logger.debug(f"Deleted temporary index file: {index_file}")
    # Return the final alignment map files.
    return bams


def figure_alignments(fq_units: list[FastqUnit], refs: set[str]):
    """ Return a ```dict``` of every expected alignment of a sample to a
    reference sequence. Check for and remove duplicates. """
    # Map each combination of a sample and reference to a FASTQ unit.
    alignments: dict[tuple[str, str], FastqUnit] = dict()
    # Keep track of any duplicate sample-reference pairs.
    duplicates: set[tuple[str, str]] = set()
    for fq_unit in fq_units:
        # Determine which references the FASTQ reads could come from.
        if fq_unit.ref is None:
            # The FASTQ contains reads from potentially all references.
            fq_refs = refs
        else:
            # The FASTQ contains reads from only one reference.
            # Confirm that the reference actually exists.
            if fq_unit.ref not in refs:
                logger.critical(
                    f"Reference '{fq_unit.ref}' of {fq_unit} not found")
                continue
            fq_refs = fq_unit.ref,
        # Add each sample-reference pair to the expected alignments.
        for ref in fq_refs:
            sample_ref = fq_unit.sample, ref
            if sample_ref in duplicates:
                # Skip the sample-reference pair if it is a duplicate.
                continue
            try:
                # Test if the sample-reference pair is already in the
                # dict of alignments. If so, then remove it.
                alignments.pop(sample_ref)
            except KeyError:
                # If not, then add the FASTQ to the dict of alignments,
                # keyed by its sample-reference pair.
                alignments[sample_ref] = fq_unit
            else:
                # If so, then flag it as a duplicate.
                logger.warning(f"Duplicate sample and reference: {sample_ref}")
                duplicates.add(sample_ref)
    # Return a duplicate-free dict of alignments.
    return alignments


def check_fqs_bams(alignments: dict[tuple[str, str], FastqUnit],
                   out_dir: Path):
    """ Return every FASTQ unit on which alignment must be run and every
    expected BAM file that already exists. """
    alignments_missing: dict[tuple[str, str], FastqUnit] = dict()
    bams_existing: list[Path] = list()
    for (sample, ref), fq_unit in alignments.items():
        # Determine the path of the BAM file expected to result from the
        # alignment of the sample to the reference.
        bam_expect = path.OneRefAlignmentOutFilePath(top=str(out_dir),
                                                     module=path.Module.ALIGN,
                                                     sample=sample,
                                                     ref=ref,
                                                     ext=path.BAM_EXT).path
        if bam_expect.is_file():
            # If the BAM file already exists, then add it to the dict of
            # BAM files that have already been aligned.
            bams_existing.append(bam_expect)
        else:
            # If at least one BAM file for a FASTQ unit does not exist,
            # then align the FASTQ.
            alignments_missing[sample, ref] = fq_unit
    return alignments_missing, bams_existing


def merge_nondemult_fqs(fq_units: Iterable[FastqUnit]):
    """ For every FASTQ that is not demultiplexed, merge all the keys
    that map to the FASTQ into one key: (sample, None). Merging ensures
    that every non-demultiplexed FASTQ is aligned only once to the whole
    set of references, not once for every reference in the set. This
    function is essentially the inverse of ```figure_alignments```. """
    merged: dict[tuple[str, str | None], FastqUnit] = dict()
    for fq_unit in fq_units:
        merged[fq_unit.sample, fq_unit.ref] = fq_unit
    return list(merged.values())


def list_fqs_bams(fq_units: list[FastqUnit],
                  refs: set[str],
                  out_dir: Path):
    """ List every FASTQ that needs to be aligned and every expected BAM
    file that already exists. """
    # Determine all possible alignments of a sample and reference.
    alignments = figure_alignments(fq_units, refs)
    # Determine which alignments need to be / have already been run.
    alignments_missing, bams_existing = check_fqs_bams(alignments, out_dir)
    # Merge entries for each non-demultiplexed FASTQ.
    fqs_to_align = merge_nondemult_fqs(alignments_missing.values())
    return fqs_to_align, set(bams_existing)


def get_bam_files(fq_units: list[FastqUnit],
                  fasta: Path, *,
                  out_dir: Path,
                  rerun: bool,
                  **kwargs) -> tuple[str, ...]:
    """ Run the alignment pipeline and return a tuple of all BAM files
    from the pipeline. """
    if rerun:
        # Rerun all alignments.
        fqs_to_align = fq_units
        bams = set()
    else:
        # Get the names of all reference sequences.
        refs = {ref for ref, _ in parse_fasta(fasta)}
        # Run only the alignments whose outputs do not yet exist.
        fqs_to_align, bams = list_fqs_bams(fq_units, refs, out_dir)
    if fqs_to_align:
        # Align all FASTQs that need to be aligned.
        bams_new = set(fqs_pipeline(fq_units, fasta, out_dir=out_dir, **kwargs))
    else:
        logger.warning("All given FASTQ files have already been aligned")
        bams_new = set()
    # Merge the existing and new BAM paths into a tuple of strings.
    return tuple(map(str, bams | bams_new))



"""
Logging configuration module of DREEM

Purpose
-------
Central manager of logging.
"""

import logging
import os
from subprocess import CompletedProcess
import sys

#import pkg_resources

#pkg_version = pkg_resources.get_distribution("dreem").version

#WELCOME = f"""
#Welcome to DREEM version {pkg_version}
#running on {sys.platform}
#with {os.cpu_count()} processors.
#"""



MAX_VERBOSE = 2
MAX_QUIET = 2
FILE_MSG_FORMAT = "LOGMSG>\t%(asctime)s\t%(name)s\t%(levelname)s\n%(message)s\n"
STREAM_MSG_FORMAT = "%(levelname)s>\t%(message)s"


LEVEL_NAME = {
    logging.DEBUG: "debug",
    logging.INFO: "info",
    logging.WARNING: "warning",
    logging.ERROR: "error",
    logging.CRITICAL: "critical",
}


def get_dreem_logger():
    """ Return the main DREEM logger. """
    if __name__ != (expect_name := "dreem.util.logs"):
        raise ValueError(
            f"{__file__} is named '{__name__}' (expected '{expect_name}')")
    dreem_logger_name = __name__.split(".")[0]
    return logging.getLogger(dreem_logger_name)


def get_verbosity(verbose: int = 0, quiet: int = 0):
    """ Get the logging level based on the verbose and quiet arguments.

    Parameters
    ----------
    verbose: int [0, 2]
        0 (): Log only warnings and errors
        1 (-v): Also log status updates
        2 (-vv): Also log detailed information (useful for debugging)
    quiet: int [0, 2]
        0 (): Suppress only status updates and detailed information
        1 (-q): Also suppress warnings
        2 (-qq): Also suppress non-critical error messages (discouraged)

    Giving both ```verbose``` and ```quiet``` flags causes the verbosity
    to default to ```verbose=0```, ```quiet=0```.
    """

    # Limit verbose and quiet to 2.
    if verbose > MAX_VERBOSE:
        logging.warning(f"Setting 'verbose' to {MAX_VERBOSE} (got {verbose})")
        verbose = MAX_VERBOSE
    if quiet > MAX_QUIET:
        logging.warning(f"Setting 'quiet' to {MAX_QUIET} (got {quiet})")
        quiet = MAX_QUIET

    # Set logging level based on verbose and quiet.
    if (verbose, quiet) == (2, 0):
        return logging.DEBUG
    if (verbose, quiet) == (1, 0):
        return logging.INFO
    if (verbose, quiet) == (0, 0):
        return logging.WARNING
    if (verbose, quiet) == (0, 1):
        return logging.ERROR
    if (verbose, quiet) == (0, 2):
        return logging.CRITICAL

    get_dreem_logger().warning(f"Invalid options: verbose={verbose}, "
                               f"quiet={quiet}. Setting both to 0")
    return get_verbosity(0, 0)


def config(verbose: int, quiet: int, log_file: str | None = None):
    """ Configure the main DREEM logger with handlers and verbosity. """
    # Set up logger.
    logger = get_dreem_logger()
    logger.setLevel(get_verbosity(verbose=MAX_VERBOSE))
    # Add stream handler.
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(get_verbosity(verbose, quiet))
    stream_handler.setFormatter(logging.Formatter(STREAM_MSG_FORMAT))
    logger.addHandler(stream_handler)
    # Add file handler.
    if log_file is not None:
        file_handler = logging.FileHandler(log_file, "a")
        file_handler.setLevel(get_verbosity(verbose=MAX_VERBOSE))
        file_handler.setFormatter(logging.Formatter(FILE_MSG_FORMAT))
        logger.addHandler(file_handler)
    #logger.info(WELCOME)


def log_process(logger: logging.Logger, process: CompletedProcess):
    """ Log the output and error messages of a process. """
    if process.stdout:
        logger.debug(f"STDOUT of {process.args}:\n{process.stdout.decode()}")
    if process.stderr:
        logger.debug(f"STDERR of {process.args}:\n{process.stderr.decode()}")



import cProfile
import os
import json

from click import Context, group, pass_context

from . import align, cluster, demultiplex, vector, aggregate, draw
from .util.dependencies import *
from .util import docdef, logs
from .util.cli import (merge_params, opt_demultiplex, opt_cluster,
                       opt_verbose, opt_quiet, opt_log, opt_profile,
                       opt_version)


logging_params = [
    opt_verbose,
    opt_quiet,
    opt_log,
    opt_profile,
]

misc_params = [
    opt_version,
]

all_params = merge_params(logging_params,
                          [opt_demultiplex],
                          demultiplex.params,
                          align.params,
                          vector.params,
                          [opt_cluster],
                          cluster.params,
                          aggregate.params,
                          aggregate.params,
                          draw.params,
                          misc_params,
                          )


# Group for all DREEM commands
@group(params=all_params,
       invoke_without_command=True,
       context_settings={"show_default": True})
@pass_context
def cli(ctx: Context, verbose: int, quiet: int, log: str, profile: str,
        **kwargs):
    """ DREEM command line interface """
    # Configure logging.
    logs.config(verbose, quiet, log_file=log)
    # If no subcommand was given, then run the entire pipeline.
    if ctx.invoked_subcommand is None:
        if profile:
            profile_path = os.path.abspath(profile)
            # Profile the program as it runs and write results to the
            # file given in the parameter profile.
            os.makedirs(os.path.dirname(profile_path), exist_ok=True)
            cProfile.runctx("run(**kwargs)",
                            globals=globals(),
                            locals=locals(),
                            filename=profile_path,
                            sort="time")
        else:
            # Run without profiling.
            run(**kwargs)


# Add all commands to the DREEM CLI command group.
# cli.add_command(test.cli)
cli.add_command(demultiplex.cli)
cli.add_command(align.cli)
cli.add_command(vector.cli)
cli.add_command(cluster.cli)
cli.add_command(aggregate.cli)
cli.add_command(draw.cli)


@docdef.auto()
def run(*,
        # General options
        out_dir: str,
        temp_dir: str,
        save_temp: bool,
        rerun: bool,
        max_procs: int,
        parallel: bool,
        library: str,
        fasta: str,
        fastqs: tuple[str],
        fastqi: tuple[str],
        fastq1: tuple[str],
        fastq2: tuple[str],
        phred_enc: int,
        # Demultiplexing options
        demulti_overwrite: bool,
        demult_on: bool,
        parallel_demultiplexing: bool,
        clipped: int,
        mismatch_tolerence: int,
        index_tolerance: int,
        barcode_start: int,
        barcode_length: int,
        # Alignment options
        fastqs_dir: tuple[str],
        fastqi_dir: tuple[str],
        fastq12_dir: tuple[str],
        fastqc: bool,
        qc_extract: bool,
        cut: bool,
        cut_q1: int,
        cut_q2: int,
        cut_g1: tuple[str],
        cut_a1: tuple[str],
        cut_g2: tuple[str],
        cut_a2: tuple[str],
        cut_o: int,
        cut_e: float,
        cut_indels: bool,
        cut_nextseq: bool,
        cut_discard_trimmed: bool,
        cut_discard_untrimmed: bool,
        cut_m: int,
        bt2_local: bool,
        bt2_discordant: bool,
        bt2_mixed: bool,
        bt2_dovetail: bool,
        bt2_contain: bool,
        bt2_unal: bool,
        bt2_score_min: str,
        bt2_i: int,
        bt2_x: int,
        bt2_gbar: int,
        bt2_l: int,
        bt2_s: str,
        bt2_d: int,
        bt2_r: int,
        bt2_dpad: int,
        bt2_orient: str,
        # Vectoring
        bamf: tuple[str],
        bamd: tuple[str],
        coords: tuple[tuple[str, int, int], ...],
        primers: tuple[tuple[str, str, str], ...],
        primer_gap: int,
        autosect: bool,
        min_phred: int,
        ambid: bool,
        strict_pairs: bool,
        batch_size: float,
        # Clustering
        clust: bool,
        mp_report: tuple[str],
        max_clusters: int,
        num_runs: int,
        signal_thresh: float,
        include_gu: bool,
        include_del: bool,
        polya_max: int,
        min_iter: int,
        max_iter: int,
        convergence_cutoff: float,
        min_reads: int,
        # Aggregation
        samples: str,
        rnastructure_path: str,
        bv_files: tuple[str],
        clustering_file: str,
        rnastructure_use_temp: bool,
        rnastructure_fold_args: str,
        rnastructure_use_dms: str,
        rnastructure_dms_min_unpaired_value: float,
        rnastructure_dms_max_paired_value: float,
        rnastructure_deltag_ensemble: bool,
        rnastructure_probability: bool,
        # Drawing
        inpt: tuple[str],
        flat: bool,
        section: str,
        mutation_fraction: bool,
        mutation_fraction_identity: bool,
        base_coverage: bool,
        mutation_per_read_per_reference: bool,
        mutations_in_barcodes: bool,
        # Misc
        version: bool
        ):

    if version:
        #print(f"DREEM version {__version__}")
        return 0

    check_bowtie2_exists()
    check_cutadapt_exists()
    check_fastqc_exists()
    check_samtools_exists()
    check_rnastructure_exists(rnastructure_path)
    
    """ Run entire DREEM pipeline. """
    # Demultiplexing
    if demult_on:
        fastqs_dir_dm, fastqi_dir_dm, fastq12_dir_dm = demultiplex.run(
            fasta=fasta,
            library=library,
            out_dir=out_dir,
            temp_dir=temp_dir,
            demulti_overwrite=demulti_overwrite,
            fastq1=fastq1,
            fastq2=fastq2,
            clipped=clipped,
            index_tolerance=index_tolerance,
            mismatch_tolerence=mismatch_tolerence,
            parallel_demultiplexing=parallel_demultiplexing,
            barcode_start=barcode_start,
            barcode_length=barcode_length,

        )

        fastqs = ()
        fastqi = ()
        fastq1 = ()
        fastq2 = ()

        fastqs_dir = fastqs_dir + fastqs_dir_dm
        fastqi_dir = fastqi_dir + fastqi_dir_dm
        fastq12_dir = fastq12_dir + fastq12_dir_dm
    # Alignment
    bamf += align.run(
        out_dir=out_dir,
        temp_dir=temp_dir,
        save_temp=save_temp,
        rerun=rerun,
        max_procs=max_procs,
        parallel=parallel,
        fasta=fasta,
        fastqs=fastqs,
        fastqi=fastqi,
        fastq1=fastq1,
        fastq2=fastq2,
        fastqs_dir=fastqs_dir,
        fastqi_dir=fastqi_dir,
        fastq12_dir=fastq12_dir,
        phred_enc=phred_enc,
        fastqc=fastqc,
        qc_extract=qc_extract,
        cut=cut,
        cut_q1=cut_q1,
        cut_q2=cut_q2,
        cut_g1=cut_g1,
        cut_a1=cut_a1,
        cut_g2=cut_g2,
        cut_a2=cut_a2,
        cut_o=cut_o,
        cut_e=cut_e,
        cut_indels=cut_indels,
        cut_nextseq=cut_nextseq,
        cut_discard_trimmed=cut_discard_trimmed,
        cut_discard_untrimmed=cut_discard_untrimmed,
        cut_m=cut_m,
        bt2_local=bt2_local,
        bt2_discordant=bt2_discordant,
        bt2_mixed=bt2_mixed,
        bt2_dovetail=bt2_dovetail,
        bt2_contain=bt2_contain,
        bt2_unal=bt2_unal,
        bt2_score_min=bt2_score_min,
        bt2_i=bt2_i,
        bt2_x=bt2_x,
        bt2_gbar=bt2_gbar,
        bt2_l=bt2_l,
        bt2_s=bt2_s,
        bt2_d=bt2_d,
        bt2_r=bt2_r,
        bt2_dpad=bt2_dpad,
        bt2_orient=bt2_orient)
    # Vectoring
    bv_files += vector.run(
        out_dir=out_dir,
        temp_dir=temp_dir,
        save_temp=save_temp,
        rerun=rerun,
        max_procs=max_procs,
        parallel=parallel,
        fasta=fasta,
        bamf=bamf,
        bamd=bamd,
        library=library,
        autosect=autosect,
        coords=coords,
        primers=primers,
        primer_gap=primer_gap,
        phred_enc=phred_enc,
        min_phred=min_phred,
        ambid=ambid,
        strict_pairs=strict_pairs,
        batch_size=batch_size,
    )
    if clust:
        cluster_results = cluster.run(
            mp_report=bv_files,
            max_clusters=max_clusters,
            min_iter=min_iter,
            max_iter=max_iter,
            signal_thresh=signal_thresh,
            include_gu=include_gu,
            include_del=include_del,
            polya_max=polya_max,
            min_reads=min_reads,
            convergence_cutoff=convergence_cutoff,
            num_runs=num_runs,
            max_procs=max_procs
        )
    else:
        cluster_results = ""
    # Aggregate
    aggregate.main.run(
        out_dir=out_dir,
        temp_dir=temp_dir,
        save_temp=save_temp,
        library=library,
        samples=samples,
        fasta=fasta,
        bv_files=bv_files,
        clustering_file=cluster_results,
        rnastructure_path=rnastructure_path,
        rnastructure_use_temp=rnastructure_use_temp,
        rnastructure_fold_args=rnastructure_fold_args,
        rnastructure_use_dms=rnastructure_use_dms,
        rnastructure_dms_min_unpaired_value=rnastructure_dms_min_unpaired_value,
        rnastructure_dms_max_paired_value=rnastructure_dms_max_paired_value,
        rnastructure_deltag_ensemble=rnastructure_deltag_ensemble,
        rnastructure_probability=rnastructure_probability,
    )

    draw.run(
        inpt=list(inpt) + [json.load(open(os.path.join(out_dir, f), 'r')) for f in os.listdir(out_dir) if
                           f.endswith(".json")],
        out_dir=out_dir,
        flat=flat,
        mutation_fraction=mutation_fraction,
        mutation_fraction_identity=mutation_fraction_identity,
        base_coverage=base_coverage,
        mutation_per_read_per_reference=mutation_per_read_per_reference,
        mutations_in_barcodes=mutations_in_barcodes,
        section=section
    )


if __name__ == "__main__":
    cli()



import os

import pandas as pd
from click import command

from ..util import docdef
from ..util.cli import (opt_out_dir, opt_draw_input,
                        opt_flat, opt_coords,opt_section, 
                        opt_mutation_fraction, opt_mutation_fraction_identity,
                        opt_base_coverage, opt_mutations_in_barcodes,
                        opt_mutations_per_read_per_sample,  
                        )
from ..util.dump import *
from .study import Study
from logging import getLogger; logger = getLogger(__name__)

params = [
    opt_draw_input,
    opt_out_dir,
    opt_flat,
    opt_coords,
    opt_section,
    opt_mutation_fraction,
    opt_mutation_fraction_identity,
    opt_base_coverage,
    opt_mutations_in_barcodes,
    opt_mutations_per_read_per_sample,
]


@command("draw", params=params)
def cli(**kwargs):
    return run(**kwargs)


@docdef.auto()
def run( 
        inpt: tuple[str], 
        *,
        flat: list,
        out_dir: str,
        coords: tuple,
        section: str,
        mutation_fraction: bool,
        mutation_fraction_identity:bool,
        base_coverage: bool,
        mutations_in_barcodes: bool,
        mutation_per_read_per_reference: bool,
        ):
    """Run the draw command.

    """

    if type(inpt) == str:
        inpt = (inpt,)
    if type(inpt[0]) == str:
        data = [json.loads(open(i).read()) for i in inpt]  
        logger.info('Loaded {} input files'.format(len(data)))
    else:
        data = inpt
    
    study = Study(
        data = data,
    )
    
    # Check output directory.
    out_dir = os.path.join(out_dir, 'draw')
    if not os.path.isdir(out_dir):
        os.makedirs(out_dir)
        logger.info('Created output directory: {}'.format(out_dir))
 
    for sample in study.get_samples():
        logger.info('Drawing sample: {}'.format(sample))
            
        path = os.path.join(out_dir, sample)
        if not os.path.isdir(path):
            os.makedirs(path)
            logger.info('Created output directory: {}'.format(path))

        refs, section, base_index = select_coords(coords, sample, study, section)
        
        all_plots = {'mutation_fraction': mutation_fraction, 'mutation_fraction_identity': mutation_fraction_identity, 'base_coverage': base_coverage, 'mutations_in_barcodes': mutations_in_barcodes, 'mutation_per_read_per_reference': mutation_per_read_per_reference}
        logger.info('Drawing {} references - section pairs'.format(len(refs)) +'\n\t'+ '\n\t'.join(['{} - {}'.format(ref, section) for ref, section in zip(refs, section)])\
                    + '\n' + 'plot types: \n\t' + '\n\t'.join([k for k, v in all_plots.items() if v])\
            )

        for ref, section, base_index in zip(refs, section, base_index):
                                
            if flat:
                prefix = os.path.join(path, '__'.join([ref, section,'']))
            else:
                prefix = os.path.join(path, ref, section , '')
                os.makedirs(prefix, exist_ok=True)
                        
            if mutation_fraction:
                study.mutation_fraction(
                    sample=sample,
                    reference=ref,
                    section=section,
                    cluster='pop_avg',
                    #base_index = base_index,
                    to_html = os.path.join(prefix,'mutation_fraction.html') if not flat else prefix + 'mutation_fraction.html',
                )
            
            if mutation_fraction_identity:
                study.mutation_fraction_identity(
                    sample=sample,
                    reference=ref,
                    section=section,
                    cluster='pop_avg',
                    #base_index = base_index,
                    to_html = os.path.join(prefix,'mutation_fraction_identity.html') if not flat else prefix + 'mutation_fraction_identity.html',
                )
            
            if base_coverage:
                study.base_coverage(
                    sample=sample,
                    reference=ref,
                    section=section,
                    cluster='pop_avg',
                    #base_index = base_index,
                    to_html = os.path.join(prefix,'base_coverage.html') if not flat else prefix + 'base_coverage.html',
                )
                
            if mutations_in_barcodes:
                study.mutations_in_barcodes(
                    sample=sample,
                    reference=ref,
                    section=section,
                    cluster='pop_avg',
                    #base_index = base_index,
                    to_html = os.path.join(prefix, 'mutations_in_barcodes.html') if not flat else prefix + 'mutations_in_barcodes.html',
                )
                
            if mutation_per_read_per_reference:
                study.mutation_per_read_per_reference(
                    sample=sample,
                    reference=ref,
                    section=section,
                    cluster='pop_avg',
                    #base_index = base_index,
                    to_html = os.path.join(prefix, 'mutation_per_read_per_reference.html') if not flat else prefix + 'mutation_per_read_per_reference.html',
                )             
                
        logger.info('Done drawing sample: {}'.format(sample))  


def select_coords(coords, sample, study, section):
    
    refs, section_out, base_index = [], [], []

    for t in coords:
        refs.append(t[0])
        section_out.append('full')
        base_index.append(np.arange(t[1], t[2] + 1))
    
    for s in section:
        for ref in study.df[study.df['sample']==sample]['reference'].unique():
            if s in study.df[(study.df['sample']==sample) & (study.df['reference']==ref)]['section'].unique():
                refs.append(ref)
                section_out.append(s)
                base_index.append(None) 

    return refs, section_out, base_index
              



"""
Path module of DREEM
========================================================================
Purpose
========================================================================
Several modules in DREEM produce files used by other modules.
For example, the alignment module creates alignment map files,
from which the vectoring module generates mutation vector files,
from which the aggregate module computes statistics and generates plots.
Modules that pass files to each other must agree on
- the path to the file, so that the second module can find the file
- the meaning of each part of the path, so that the second module can
  parse information contained in the path
Although these path conventions could be written separately in each
module this strategy is not ideal for several reasons:
- It would risk inconsistencies among the modules, causing bugs.
- Changing the conventions would require modifying every module,
  which would be not only tedious but also risky for the first reason.
- Defining all the conventions in one place would reduce the size of the
  code base, improving readability, maintainability, and distribution.  
This module defines all file path conventions for all other modules.
Concepts
========================================================================
Paths and path segments
-----------------------
Every path is represented as a collection of path segments.
Each segment is the space between two path separators, except for the
first ("top") segment of a path, which can contain separators.
For example, this path to an alignment map file
```/home/rfranklin/tmv/exp58/output/alignment/dms2/tmv-genome.bam```
would be represented as the following segments:
- top (full path of the top-level directory): ```/home/rfranklin/tmv/exp58/```
- module (DREEM module producing the results): ```alignment```
- sample (name of the sample from the experiment): ```dms2```
- ref (name of the reference sequence the sample was aligned to): tmv-genome
Classes of paths
----------------
Usage
========================================================================
Creating a path of a specific type using its class
--------------------------------------------------
An instance of a specific type of path can be created in three ways:
1. Calling the class directly, giving the names and values of the path
segments as keyword arguments:
>>> xam_inp = OneRefAlignmentOutFilePath(top=os.getcwd(),
...                                       module=Module.ALIGN,
...                                       sample="dms2",
...                                       ref="tmv-rna",
...                                       ext=".bam")
>>> assert str(xam_inp) == os.getcwd() + "/alignment/dms2/tmv-rna.bam"
2. Calling the class directly, giving the names and values of the path
segments as a dictionary of keyword arguments:
>>> bam_fields = {"top": os.getcwd(), "module": Module.ALIGN,
...               "sample": "dms2", "ref": "tmv-rna", "ext": ".bam"}
>>> xam_inp = OneRefAlignmentOutFilePath(**bam_fields)
>>> assert str(xam_inp) == os.getcwd() + "/alignment/dms2/tmv-rna.bam"
3. Parsing the path from a string (or from any other object whose
   __str__ method returns a valid path, such as a pathlib.Path object):
>>> path = os.getcwd() + "/alignment/dms2/tmv-rna.bam"
>>> xam_inp = OneRefAlignmentOutFilePath.parse(path)
>>> assert xam_inp.dict() == {"top": os.getcwd(),
...                            "module": Module.ALIGN,
...                            "sample": "dms2",
...                            "ref": "tmv-rna",
...                            "ext": ".bam"}
Creating a path by inferring the type from the fields
-----------------------------------------------------
This module also provides a convenience function called make_path, which
behaves identically to the first two examples (of calling the class directly),
except that instead of the code specifying the type of path to be created,
make_path infers the type from the keywords arguments. For example:
A few notes on the function make_path:
- There is no equivalent of parse_path() for make_path because it is impossible
  to uniquely assign fields to segments of an arbitrary path if neither the
  structure nor the values of the fields are known.
- A TypeError is raised if there is no class of path whose field names match the
  names of the keyword arguments given to make_path.
- In most cases, this function suffices and alleviates the need to import every
  class of path that will be used in a module.
Implementation Details
========================================================================
Representation of paths and path segments
-----------------------------------------
In this module, every path is represented as a sequence of path segments. Every
kind of path is represented by a subclass of BasePath, and every kind of segment
is represented by a subclass of BaseSeg. For example, one subclass of BasePath
represents the full path to a mutation vector report file, and is made of path
segments that represent the name of the report file, the name of the directory
in which it resides, and so on up.
Responsibilities of paths and path segments
-------------------------------------------
Each path segment class is responsible for storing information contained in the
segment (i.e. the value of each field), generating a string representation of
the segment from the values of the fields, and parsing string representations
of the segment to determine the values of the fields encoded within. Each path
class is responsible for storing a series of path segments in the proper order,
ensuring that the path has all expected and no unexpected segments, ensuring
that the values of all of its segments are consistent with each other, and
generating and parsing string representations of the entire path by calling the
corresponding method of each segment and assembling the results.
Path segment classes and inheritance structure
----------------------------------------------
Every path segment class represents a specific component of a path, such as a
directory named after a sample or the name of a mutation vector batch file.
The base class of every path segment is BaseSeg, which does not represent any
specific path segment and should not be instantiated. Every other type of path
segment is a subclass of BaseSeg. BaseSeg is itself decorated with @dataclass
but contains no fields. However, it provides all shared methods for path segment
subclasses (all of which are also decorated with @dataclass):
- Initialization: __init__ (from @dataclass), __post_init__, validate
- Field access: fields, field_names, astuple, asdict, replace, _get_dtypes
- Interface: parse_seg, format_seg
The subclasses of BaseSeg that represent directories are structured as follows:
Subclass of BaseSeg     Fields      Constraints
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BaseSeg                 -           -
    TopSeg              top         must have permission to read & write
    SubSeg              -           may not contain a path separator
        ModSeg          module      must be a valid module name
        StepStepSeg     step        must be a valid step name
        SampleSeg       sample      -
        RefsetSeg       refset      -
        RefSeg          ref         -
        StructSeg       -           -
            SectionSeg   end5, end3  1 ≤ end5 (int) ≤ end3 (int)
            FileSeg     ext         must be a valid file extension
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Most segments for file types inherit from both ExtenSeg and another class that
provides one or more fields that are part of the file name:
Subclass of ExtenSeg    Also subclass of
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
MutVectorBatchSeg       -
MutVectorReportSeg	    SectionSeg
XamSeg                  -
    XamMixedSeg         RefsetSeg
    XamSplitSeg         RefSeg
FastaSeg	            -
    FastaMixedSeg	    RefsetSeg
    FastaSplitSeg	    RefSeg
FastqSeg                -
    FastqSampleSeg      SampleSeg
    FastqDemultSeg      RefSeg
Fastq1Seg               -
    Fastq1SampleSeg     SampleSeg
    Fastq1DemultSeg     RefSeg
Fastq2Seg               -
    Fastq2SampleSeg     SampleSeg
    Fastq2DemultSeg     RefSeg
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Implementation of path and path segment classes as Pydantic models
------------------------------------------------------------------
"""

# Imports ##############################################################

from __future__ import annotations
from enum import Enum
from functools import cache
from inspect import getmembers, isclass
import itertools
import os
import pathlib
import re
from string import ascii_letters, digits
import sys
from typing import Any, ClassVar, Iterable

from pydantic import BaseModel, Extra, NonNegativeInt, PositiveInt
from pydantic import root_validator, validator


# Constants ############################################################

# Valid/invalid characters in fields
VALID_CHARS = ascii_letters + digits + "_~=+-"
VALID_CHARS_SET = set(VALID_CHARS)
VALID_FIELD_PATTERN = f"([{VALID_CHARS}]+)"
VALID_FIELD_REGEX = re.compile(f"^{VALID_FIELD_PATTERN}$")

TOP_KEY = "top"
EXT_KEY = "ext"
EXTS_KEY = "exts"


class Module(Enum):
    DEMULT = "demultiplexing"
    ALIGN = "alignment"
    VECTOR = "vectoring"
    CLUSTER = "clustering"
    AGGREG = "aggregation"


class Step(Enum):
    ALIGN_REFS = "align-0_refs"
    ALIGN_TRIM = "align-1_trim"
    ALIGN_ALIGN = "align-2_align"
    ALIGN_DEDUP = "align-3_dedup"
    ALIGN_SORT = "align-4_sort"
    ALIGN_SPLIT = "align-5_split"
    VECTOR_BAMS = "vector-0_bams"
    VECTOR_SELECT = "vector-1_select"
    VECTOR_SORT = "vector-2_sort"


class Fastqc(Enum):
    QC_INPUT = "qc-inp"
    QC_TRIM = "qc-cut"


# File extensions

EXT_PATTERN = "([.].+)"
FASTA_EXTS = (".fasta", ".fna", ".fa")
BOWTIE2_INDEX_EXTS = (".1.bt2", ".2.bt2", ".3.bt2", ".4.bt2",
                      ".rev.1.bt2", ".rev.2.bt2")
BOWTIE2_INDEX_PATTERN = "([.][1-4][.]bt2|[.]rev[.][1-2][.]bt2)"
FQ_EXTS = (".fastq", ".fq", ".fastq.gz", ".fq.gz",
           "_001.fastq", "_001.fq", "_001.fastq.gz", "_001.fq.gz")
FQ_PAIRED_EXTS_TEMPLATES = ("_R{}{}", "_mate{}{}", "_{}_sequence{}")
FQ1_EXTS = tuple(template.format(1, ext) for template, ext in
                 itertools.product(FQ_PAIRED_EXTS_TEMPLATES, FQ_EXTS))
FQ2_EXTS = tuple(template.format(2, ext) for template, ext in
                 itertools.product(FQ_PAIRED_EXTS_TEMPLATES, FQ_EXTS))
SAM_EXT = ".sam"
BAM_EXT = ".bam"
CRAM_EXT = ".cram"
XAM_EXTS = (SAM_EXT, BAM_EXT, CRAM_EXT)
BAI_EXT = f"{BAM_EXT}.bai"
XAI_EXTS = (BAI_EXT,)


# Exceptions ###########################################################

class PathError(Exception):
    """ Any error involving a path """


class PathTypeError(PathError, TypeError):
    """ Use of the wrong type of path or segment """


class PathValueError(PathError, ValueError):
    """ Invalid value of a path segment field """


# Path functions #######################################################

def sanitize(path: Any):
    return os.path.realpath(os.path.normpath(os.path.abspath(str(path))))


# Path segment classes #################################################

class BaseSeg(BaseModel):
    """ Abstract base class for all classes that represent segments of a path.
    Should not be instantiated. """

    class Config:
        extra = Extra.forbid

    @classmethod
    def fields(cls):
        return cls.__fields__

    @classmethod
    def n_fields(cls):
        return len(cls.fields())

    @classmethod
    def keys(cls):
        return list(cls.fields().keys())

    @classmethod
    def build_dict(cls, values: Iterable):
        return dict(zip(cls.keys(), values, strict=True))

    def values(self):
        return list(self.dict().values())

    @classmethod
    def parse(cls, segment: Any):
        """ Return a new instance of the class given a segment as a str-like
        argument. Non-structured segments (including this base class) must have
        exactly one field. """
        return cls(**cls.build_dict([segment]))

    def __str__(self):
        """ Return a string representation of the first field of the
        segment. Non-structured segments (including this base class)
        must have exactly one field. """
        try:
            value, = self.values()
        except ValueError:
            raise PathTypeError(f"str is undefined for {self.__class__} "
                                f"with {self.n_fields()} (≠ 1) fields.")
        return str(value)


class TopSeg(BaseSeg):
    """ Top-level working directory of DREEM. """
    top: str

    @validator(TOP_KEY)
    def sanitize_top_dir(cls, top: Any):
        return sanitize(top)


class SubSeg(BaseSeg):
    """ Any path level below the top level. """

    @root_validator(pre=True)
    def fields_must_match_valid_field_regex(cls, values: dict[str, Any]):
        for key, value in values.items():
            if isinstance(value, Enum):
                # For Enum types, the value must be obtained from the
                # value attribute, because simply calling str(value) on
                # the original type will yield the string {NAME}.{value}
                value = value.value
            if (key != TOP_KEY and key != EXT_KEY
                    and not VALID_FIELD_REGEX.match(str(value))):
                raise PathValueError(f"{cls} got invalid '{key}' value: '{value}'")
        return values


class SampleSeg(SubSeg):
    """ Segment for a sample. """
    sample: str


class RefsetSeg(SubSeg):
    """ Segment for a set of reference sequences. """
    refset: str


class OneRefSeg(SubSeg):
    """ Segment for one reference sequence. """
    ref: str


class ModSeg(SubSeg):
    """ Segment for the outputs of a module. """
    module: Module

    def __str__(self):
        return str(self.module.value)


class StepSeg(SubSeg):
    """ Segment for the temporary files of a step. """
    step: Step

    def __str__(self):
        return str(self.step.value)


class FastqcSeg(SubSeg):
    """ Segment for a directory of FASTQC reports. """
    fastqc: Fastqc

    def __str__(self):
        return str(self.fastqc.value)


class BatchSeg(SubSeg):
    """ Segment for a batch. """
    batch: NonNegativeInt


class StructSeg(SubSeg):
    """
    Segment representing any path segment whose name contains an internal
    structure with one or more fields. For example, a segment could encode the
    name of a sample and the year, month, and day on which it was generated.
    This base class provides all the methods of structured segments but does
    not itself have any fields.
    Class attributes
    ----------------
    format_str: str
        Specify how to represent the path segment as a string, using Python's
        string formatting mini-language. The string is created by calling
        ```self.format_str.format(*self.dict())```.
        Note: ```format_str``` must be compatible with ```pattern_str```
    pattern_str: str
        Specify how to parse a string representing the path to determine the
        values of the fields. The string is parsed using ```re.match```,
        and ```match.groups()``` is used to get the values of the fields.
        Note: ```pattern_str``` must be compatible with ```format_str```
    Examples
    --------
    >>> class SampleBatchSeg(StructSeg, BatchSeg, SampleSeg):
    ...     format_str = "samp-{}_batch-{}"
    ...     pattern_str = f"samp-{VALID_FIELD_PATTERN}_batch-([0-9]+)"
    >>> str(SampleBatchSeg(sample="tmv", batch=37))
    'samp-tmv_batch-37'
    """

    format_str: ClassVar[str] = ""
    pattern_str: ClassVar[str] = ""

    @classmethod
    def _parse(cls, seg_str: Any):
        """ Return a new instance of the class by parsing a ```str```-like
        ```segstr``` with ```re.match``` (using ```pattern_str``` as the
        regular expression pattern), then passing the parsed groups to the
        class constructor. """
        # Try to parse segstr with re.match, using pattern_str as the pattern.
        if match := re.match(cls.pattern_str, str(seg_str)):
            # If segstr matched the pattern, then create keyword arguments from
            # the matched values, and use them to initialize a new instance.
            return cls(**cls.build_dict(match.groups()))
        # If segstr did not match the pattern, then the parsing failed.
        raise PathValueError(
            f"Segment '{seg_str}' failed to match pattern {cls.pattern_str}")

    @classmethod
    def parse(cls, seg_str: Any):
        """
        Return a new instance of the class by parsing the values of the
        fields from a string and passing them to the class constructor.
        Parameters
        ----------
        seg_str: Any
            Literal representation of the segment as a ```str``` (or as
            any other object ```x``` for which ```str(x)``` is a valid
            path segment, e.g. an instance of ```pathlib.Path```).
        
        Returns
        -------
        StructSeg
            New instance of StructSeg with fields from ```seg_str```;
            satisfies ```str(seg_inst) == str(seg_str)```.
        Raises
        ------
        ValueError
            if the string representation of the newly created instance
            of the class did not match ```seg_str```.
        """
        # Create a new instance of the class by parsing seg_str.
        seg_inst = cls._parse(seg_str)
        # Confirm that formatting seg_inst yields the original seg_str.
        if (new_str := seg_inst._format()) != str(seg_str):
            # If not, an error occurred during parsing.
            raise PathValueError(
                f"The new instance was formatted as '{new_str}'"
                f"(failed to match input '{seg_str}')")
        # If so, the new instance was parsed correctly.
        return seg_inst

    def _format(self):
        """ Return the result of calling the ```format``` method of
        ```format_str``` with the field values as arguments. """
        return self.format_str.format(*self.values())

    def __str__(self) -> str:
        """
        Return a string representation of the path segment by inserting the
        values of the segment's fields into its format specification.
        """
        # Create the formatted string.
        seg_str = self._format()
        # Try to parse that string into a new instance, then confirm that the
        # fields of the new instance match those of the self instance.
        if (parse := self._parse(seg_str)) != self:
            # Raise an error if the fields did not match.
            raise PathValueError(
                f"String representation '{seg_str}' was parsed as "
                f"{repr(parse)} (failed to match {repr(self)})")
        # If the fields match, then the formatted string is valid.
        return seg_str


class SectionSeg(StructSeg):
    """
    Segment for a directory of a section of a reference sequence.
    Fields
    ------
    end5: int
        The 5'-most coordinate in the section; 1-indexed, inclusive.
    end3: int
        The 3'-most coordinate in the section; 1-indexed, inclusive.
    """

    end5: PositiveInt
    end3: PositiveInt

    format_str = "{}-{}"
    pattern_str = "([0-9]+)-([0-9]+)"

    @root_validator()
    def end5_le_end3(cls, values):
        """ Validate that end5 ≤ end3 """
        if values["end5"] > values["end3"]:
            raise PathValueError(f"Got end5 ({values['end5']}) "
                                 f"> end3 ({values['end3']})")
        return values


class FileSeg(StructSeg):
    """
    Segment representing a file name with an extension.
    Fields
    ------
    ext: str
        The file extension. Must begin with a period.
    Class attributes
    ----------------
    exts: tuple[str]
        All valid file extensions for the class. Each must begin with a period.
    """
    ext: str
    exts: ClassVar[tuple[str, ...]] = tuple()

    format_str = "{}{}"
    pattern_str = f"^{VALID_FIELD_PATTERN}{EXT_PATTERN}$"

    @validator(EXT_KEY)
    def valid_file_extension(cls, ext):
        """ Validate the file extension (```ext```). It must be an
        element of the class attribute ```exts```. """
        if ext not in cls.exts:
            raise PathValueError(f"Invalid extension for {cls}: '{ext}'")
        return ext

    @classmethod
    def keys_no_ext(cls):
        """ Return a list of field keys without the 'ext' field. """
        # Create a new list of the field keys.
        keys = cls.keys()
        # Remove 'ext' from the list.
        keys.remove(EXT_KEY)
        # Return the list of keys without 'ext'.
        return keys

    @classmethod
    def _parse(cls, segstr: Any):
        if all(ext[0] not in VALID_CHARS_SET for ext in cls.exts):
            # If the extension starts with a character that is not valid if it
            # occurs in a field (e.g. the most common starting character '.'),
            # then the extension can be distinguished from the other fields by
            # the regular expression defined in the superclass.
            return super()._parse(segstr)
        # If the extension starts with a character that is valid if it occurs
        # in a field (e.g. '_R1.fq' is such an extension because it starts
        # with '_', which may occur in a field), then we need to match the
        # segment minus the extension to the pattern minus its extension.
        segstr_ = str(segstr)
        # Remove the final group, denoted by (), from the pattern string. This
        # group captures the extension, so it is removed.
        pattern_trunc = cls.pattern_str[:cls.pattern_str.rindex("(")]
        for ext in cls.exts:
            # Find the extension in exts that segstr ends with.
            if segstr_.endswith(ext):
                # Remove that extension from segstr.
                segstr_trunc = segstr_[:-len(ext)]
                # See if the truncated segment matches the truncated pattern.
                if match := re.match(pattern_trunc, segstr_trunc):
                    # If the truncated strings matched, then add the extension
                    # back to the end of the matched groups; we already know it
                    # matched because segstr_.endswith(ext) returned True.
                    return cls(**cls.build_dict(match.groups() + (ext,)))
        # If segstr did not match any extension, or did match an extension but
        # failed to match any patterns after removing the extension, then the
        # parsing failed.
        raise PathValueError(
            f"Segment '{segstr}' failed to match pattern "
            f"'{cls.pattern_str}' with any extension ({cls.exts})")


class MutVectorReportFileSeg(FileSeg):
    """ Segment for a mutation vector report file. """
    format_str = "report{}"
    pattern_str = f"report{EXT_PATTERN}"
    exts = (".json",)


class MutVectorBatchFileSeg(FileSeg, BatchSeg):
    """ Segment for a mutation vector batch file. """
    format_str = "{}{}"
    pattern_str = "([0-9]+)" + EXT_PATTERN
    exts = (".orc",)


class AbstractRefFileSeg(FileSeg):
    """ Abstract segment for a FASTA file. Do not instantiate. """
    exts = FASTA_EXTS


class RefsetSeqFileSeg(AbstractRefFileSeg, RefsetSeg):
    """ Segment for a FASTA file containing one or more references. The
    refset field is the name of the file, minus the file extension. """


class OneRefSeqFileSeg(AbstractRefFileSeg, OneRefSeg):
    """ Segment for a FASTA file that contains exactly one reference
    sequence. The ref field is the reference name (on the first line,
    after '>'). """


class AbstractBowtie2IndexFileSeg(FileSeg):
    """ Abstract segment for a Bowtie2 index. Do not instantiate. """
    exts = BOWTIE2_INDEX_EXTS
    pattern_str = f"^{VALID_FIELD_PATTERN}{BOWTIE2_INDEX_PATTERN}$"


class RefsetBowtie2IndexFileSeg(AbstractBowtie2IndexFileSeg, RefsetSeg):
    """ Segment for a Bowtie2 index of a FASTA file containing one or
    more references. """


class OneRefBowtie2IndexFileSeg(AbstractBowtie2IndexFileSeg, OneRefSeg):
    """ Segment for a Bowtie2 index of a FASTA file that contains
    exactly one reference sequence. """


class AbstractReadsFileSeg(FileSeg):
    """ Abstract segment for a standalone (single-end or paired-
    interleaved) FASTQ file. Do not instantiate. """
    exts = FQ_EXTS


class SampleReadsFileSeg(AbstractReadsFileSeg, SampleSeg):
    """ Segment for a standalone FASTQ file containing reads from one
    sample. """


class OneRefReadsFileSeg(AbstractReadsFileSeg, OneRefSeg):
    """ Segment for a standalone FASTQ file containing reads from one
    construct (i.e. a synthetic, barcoded reference) in one sample. """


class AbstractReads1FileSeg(FileSeg):
    """ Abstract segment for a paired-end FASTQ file containing mate-1 reads.
    Should not be instantiated. """
    exts = FQ1_EXTS


class SampleReads1FileSeg(AbstractReads1FileSeg, SampleSeg):
    """ Segment for a paired-end FASTQ file containing all mate-1 reads from a
    sample. """


class OneRefReads1FileSeg(AbstractReads1FileSeg, OneRefSeg):
    """ Segment for a paired-end FASTQ file containing all mate-1 reads from one
    reference (i.e. reference sequence) in one sample. """


class AbstractReads2FileSeg(FileSeg):
    """ Abstract segment for a paired-end FASTQ file containing mate-2 reads.
    Should not be instantiated. """
    exts = FQ2_EXTS


class SampleReads2FileSeg(AbstractReads2FileSeg, SampleSeg):
    """ Segment for a paired-end FASTQ file containing all mate-2 reads from a
    sample. """


class OneRefReads2FileSeg(AbstractReads2FileSeg, OneRefSeg):
    """ Segment for a paired-end FASTQ file containing all mate-2 reads
    from one construct (i.e. reference sequence) in one sample. """


class AbstractAlignmentFileSeg(FileSeg):
    """ Abstract segment for alignment map files (.sam, .bam, .cram) """
    exts = XAM_EXTS


class RefsetAlignmentFileSeg(AbstractAlignmentFileSeg, RefsetSeg):
    """ Segment for an ailgnment map file resulting from aligning a set
    of reads to a set of references """


class OneRefAlignmentFileSeg(AbstractAlignmentFileSeg, OneRefSeg):
    """ Segment for an alignment map file produced by splitting another
    alignment map file into one file for each reference """


class SectionAlignmentFileSeg(AbstractAlignmentFileSeg, SectionSeg):
    """ Segment for an alignment map file produced by taking a subset
    of a particular section from another alignment map file """
    format_str = "{}-{}{}"
    pattern_str = f"([0-9]+)-([0-9]+){EXT_PATTERN}"


class AbstractAlignmentIndexFileSeg(FileSeg):
    """ Abstract segment for index files of alignment maps """
    exts = XAI_EXTS


class RefsetAlignmentIndexFileSeg(AbstractAlignmentIndexFileSeg, RefsetSeg):
    """ Segment for an index file of an ailgnment map file resulting
    from aligning a set of reads to a set of references """


class OneRefAlignmentIndexFileSeg(AbstractAlignmentIndexFileSeg, OneRefSeg):
    """ Segment for an index file of an ailgnment map file resulting
    from aligning a set of reads to a set of references """


# Path classes #########################################################

class BasePath(BaseModel):
    """ Abstract base class for all classes that represent full,
    absolute paths. Do not instantiate. """

    @classmethod
    def fields(cls):
        return cls.__fields__

    @classmethod
    def keys(cls):
        return list(cls.fields().keys())

    @classmethod
    def _get_segment_types(cls):
        """ Return a list of the type of every segment in the path. """
        seg_types: list[type[BaseSeg]] = list()
        # Search the base classes for path segments.
        for base in cls.__bases__:
            if issubclass(base, BasePath):
                # The base class represents a path and thus also has
                # base classes of path segments: add them to the list.
                seg_types.extend(base._get_segment_types())
            elif issubclass(base, BaseSeg):
                # The base class is a path segment but not a path class.
                # Thus, it should be in the list of segments.
                seg_types.append(base)
            else:
                # Otherwise, the base class contains no path segments.
                pass
        return seg_types

    @classmethod
    def segment_types(cls) -> tuple[type[BaseSeg], ...]:
        """ Return a tuple of the type of every segment in the path, in
        order from the beginning to the end of the path. """
        seg_types = tuple(cls._get_segment_types())
        if seg_types and seg_types[0] is not TopSeg:
            raise PathTypeError(f"{cls} begins with {seg_types[0].__name__}, "
                                f"not {TopSeg.__name__}")
        return seg_types

    @classmethod
    def _parse_segments(cls,
                        remaining_seg_types: list[type[BaseSeg]],
                        path: str):
        """
        Return a dict of the names and values of the fields encoded
        within a string representation of a path.
        Parameters
        ----------
        remaining_seg_types: list[type[BaseSeg]]
            Types of segments in the path that have yet to be used to
            parse part of the path string. The segment types are ordered
            from beginning (left) to end (right) of the path. The first
            item (left-most segment type) must be ```TopSeg```.
        path: str
            Path to be parsed, represented as a string.
        Return
        ------
        dict[str, Any]
            A dict of the names (keys) and values (values) of the fields
            encoded in the path string.
        Raises
        ------
        PathTypeError
            if any part of the path remains to be parsed after all the
            types of segments have been consumed.
        """
        try:
            # Get the last (right-most) type of segment in the path that
            # has not been used to parse a segment of the path string.
            rightmost_seg_type = remaining_seg_types.pop()
        except IndexError:
            # Base case of this recursive function. The pop() operation
            # raised an IndexError because seg_types is empty. Thus, no
            # segment types remain.
            if path:
                # Any part of the path that has not yet been parsed
                # cannot be parsed, since no segment types are left.
                raise PathTypeError(f"No segments remain to parse '{path}'")
            # Return a dict with no fields, signifying that nothing
            # remains to be parsed.
            return dict()
        # At least one more segment type remains.
        if remaining_seg_types:
            # If any segment types will remain after the one that is
            # currently the right-most segment type, then parse only the
            # last (right-most) segment of the path string (i.e. after
            # the final path separator) using the right-most segment
            # type. The remaining segments (towards the left side of the
            # path) will be parsed with the remaining segment types.
            remaining_segs, rightmost_seg = os.path.split(path)
        else:
            # If the segment type that is currently the right-most type
            # is the last one remaining, then use it to parse the entire
            # remaining part of the path. To do so, consider all that
            # remains of the path to be the "right-most segment" (which
            # will be parsed by the right-most segment type), and thus
            # assign an empty string to the part of the path that will
            # remain after parsing this "right-most segment".
            remaining_segs, rightmost_seg = "", path
        # Ensure that part of the path still remains to be parsed.
        if not path:
            raise PathValueError("No part of the path remains to be parsed by "
                                 f"segment '{rightmost_seg_type.__class__}'")
        # Parse the right-most segment with the right-most segment type,
        # and parse the remaining segments with the remaining segment
        # types. This step will also catch any duplicate fields (which
        # are not supposed to occur) by raising a TypeError.
        return dict(**cls._parse_segments(remaining_seg_types, remaining_segs),
                    **rightmost_seg_type.parse(rightmost_seg).dict())

    @classmethod
    def parse(cls, path: Any):
        """
        Return a new instance of the class by parsing a given path.
        Parameters
        ----------
        path: Any
            A ```str``` representing the path to be parsed (or any other
            object ```x``` for which ```str(x)``` returns a valid path,
            e.g. an instance of ```pathlib.Path```).
        Returns
        -------
        BasePath
            A new instance of this class that represents the given path.
        Raises
        ------
        PathValueError
            if the newly created instance of the class yields a string
            representation that does not match the ```path``` argument.
        """
        path_inst = cls(**cls._parse_segments(list(cls.segment_types()),
                                              str(path)))
        if str(path_inst) != os.path.abspath(path):
            raise PathValueError(
                f"String representation of new path '{path_inst}' "
                f"failed to match input '{os.path.abspath(path)}'")
        return path_inst

    @property
    def segments(self) -> tuple[BaseSeg, ...]:
        """ Return a tuple of an instance of every segment in the path,
        in order from the beginning to the end of the path. """
        return tuple(seg_type(**{key: getattr(self, key)
                                 for key in seg_type.keys()})
                     for seg_type in self.segment_types())

    @property
    def path(self):
        """ Return a ```pathlib.Path``` instance for this path. """
        # Using @functools.cached_property instead of @functools.cache
        # would be more efficient but causes errors when applied to
        # methods of Pydantic models.
        return pathlib.Path(*map(str, self.segments))

    @property
    def parent(self):
        """ Return the parent directory of the path. """
        parent_fields = dict()
        for seg in self.segments[:-1]:
            parent_fields.update(seg.dict())
        return create(**parent_fields)

    def _get_new_fields(self, **fields):
        """ Computes the new fields obtained by merging given fields
        with the existing fields of this path. If a field occurs in both
        the given and the existing fields, the given field replaces the
        existing field. """
        return {**self.dict(), **fields}

    def edit(self, **fields):
        """ Return a new path instance based on this instance with some
        fields modified according to the keyword arguments. Set a field
        to ```None``` to delete it from the new path instance.
        Return
        ------
        BasePath
            Path: modified; Type: modified
        """
        return create(**self._get_new_fields(**fields))

    def cast(self, **fields):
        """ Return a new path instance that evaluates to the same path
        string and has been cast to a new type. The new type is found by
        merging the given fields with the existing fields of this path
        and determining which type matches those fields. Set a field to
        ```None``` to delete it from the new path instance. Note that,
        except for 'ext' (which determines file types), the values of
        the given fields do not affect the return value, since the type
        of path does not depend on the values of the fields, but rather
        on which fields it has (again, except for the field 'ext').
        Return
        ------
        BasePath
            Path: preserved; Type: modified
        """
        new_type = get_path_class_by_fields(**self._get_new_fields(**fields))
        return new_type.parse(self)

    def move(self, **fields):
        """ Return a new path instance that has the same type as the
        original but has a different path. Determine the new path based
        on the fields given as keyword arguments.
        Return
        ------
        BasePath
            Path: modified; Type: preserved
        """
        # First, create the new path (modified path and modified type)
        # using the edit method.
        new_path = self.edit(**fields)
        # Compare the types directly here rather than using isinstance
        # so that subclasses are not considered equivalent to their
        # parent classes. See the comments in __eq__ for more details.
        if new_path.__class__ is self.__class__:
            return new_path
        # If, as usually happens, the type of the new path does not
        # match the type of the previous path, then parse the new path
        # with the type of the previous path.
        return self.__class__.parse(new_path)

    def __str__(self):
        return str(self.path)

    def __eq__(self, other: Any):
        # Note that this comparison avoids using isinstance because a
        # subclass should not evaluate as equal to its parent class even
        # if the path strings are identical. This is because, unlike in
        # most cases where a subclass is a "more specific" version of
        # its parent class, in this module, object inheritance is used
        # to create the hierarchical structure of directories and files,
        # wherein a subclass in the code equates to a child directory in
        # the file system. Since a child directory or file is not just a
        # "more specific" kind of its parent directory, subclasses and
        # parent classes are considered different for equality test_input.
        return self.__class__ is other.__class__ and str(self) == str(other)


# General directory paths

class TopDirPath(BasePath, TopSeg):
    pass


class ModuleDirPath(TopDirPath, ModSeg):
    pass


class StepDirPath(ModuleDirPath, StepSeg):
    pass


# Sample, Reference, and Section directory paths

class AbstractSampleDirPath(BasePath, SampleSeg):
    """ Abstract directory named after a sample """


class SampleInDirPath(TopDirPath, AbstractSampleDirPath):
    """ Directory of input files, named after a sample """


class SampleStepDirPath(StepDirPath, AbstractSampleDirPath):
    """ Directory of temporary files, named after a sample """


class SampleOutDirPath(ModuleDirPath, AbstractSampleDirPath):
    """ Directory of output files, named after a sample """


class AbstractRefDirPath(BasePath, OneRefSeg):
    """ Abstract directory named after a reference """


class RefInDirPath(SampleInDirPath, AbstractRefDirPath):
    """ Directory of input files, named after a reference """


class RefStepDirPath(SampleStepDirPath, AbstractRefDirPath):
    """ Directory of temporary files, named after a reference """


class RefOutDirPath(SampleOutDirPath, AbstractRefDirPath):
    """ Directory of output files, named after a reference """


class AbstractSectionDirPath(BasePath, SectionSeg):
    """ Abstract directory named after a section """


class SectionInDirPath(RefInDirPath, AbstractSectionDirPath):
    """ Directory of input files, named after a section """


class SectionStepDirPath(RefStepDirPath, AbstractSectionDirPath):
    """ Directory of temporary files, named after a section """


class SectionOutDirPath(RefOutDirPath, AbstractSectionDirPath):
    """ Directory of output files, named after a section """


# Reference sequence (FASTA) and Bowtie2 index file paths

class AbstractRefsetSeqFilePath(BasePath, RefsetSeqFileSeg):
    """ Abstract FASTA file named after a set of references """


class RefsetSeqInFilePath(TopDirPath, AbstractRefsetSeqFilePath):
    """ Input FASTA file named after a set of references """


class RefsetSeqTempFilePath(StepDirPath, AbstractRefsetSeqFilePath):
    """ Temporary FASTA file named after a set of references """


class RefsetSeqOutFilePath(ModuleDirPath, AbstractRefsetSeqFilePath):
    """ Output FASTA file named after a set of references """


class AbstractRefsetBowtie2IndexFilePath(BasePath, RefsetBowtie2IndexFileSeg):
    """ Abstract Bowtie2 index file for a FASTA file named after a set
    of references """


class RefsetBowtie2IndexInFilePath(TopDirPath,
                                   AbstractRefsetBowtie2IndexFilePath):
    """ Bowtie2 index file for an input FASTA file named after a set of
    references """


class RefsetBowtie2IndexTempFilePath(StepDirPath,
                                     AbstractRefsetBowtie2IndexFilePath):
    """ Bowtie2 index file for a temporary FASTA file named after a set
    of references """


class RefsetBowtie2IndexOutFilePath(ModuleDirPath,
                                    AbstractRefsetBowtie2IndexFilePath):
    """ Bowtie2 index file for an output FASTA file named after a set of
    references """


class AbstractRefsetBowtie2IndexPrefix(BasePath, RefsetSeg):
    """ Abstract Bowtie2 index prefix for a FASTA file named after a set
    of references """


class RefsetBowtie2IndexInPrefix(TopDirPath,
                                 AbstractRefsetBowtie2IndexPrefix):
    """ Bowtie2 index prefix for an input FASTA file named after a set
    of references """


class RefsetBowtie2IndexTempPrefix(StepDirPath,
                                   AbstractRefsetBowtie2IndexPrefix):
    """ Bowtie2 index prefix for a temporary FASTA file named after a
    set of references """


class RefsetBowtie2IndexOutPrefix(ModuleDirPath,
                                  AbstractRefsetBowtie2IndexPrefix):
    """ Bowtie2 index prefix for an output FASTA file named after a set
    of references """


class AbstractOneRefSeqFilePath(BasePath, OneRefSeqFileSeg):
    """ Abstract FASTA file named after one reference """


class OneRefSeqInFilePath(TopDirPath, AbstractOneRefSeqFilePath):
    """ Input FASTA file named after one reference """


class OneRefSeqTempFilePath(StepDirPath, AbstractOneRefSeqFilePath):
    """ Temporary FASTA file named after one reference """


class OneRefSeqOutFilePath(ModuleDirPath, AbstractOneRefSeqFilePath):
    """ Output FASTA file named after one reference """


class AbstractOneRefBowtie2IndexFilePath(BasePath, OneRefBowtie2IndexFileSeg):
    """ Abstract Bowtie2 index file for a FASTA file named after one
    reference """


class OneRefBowtie2IndexInFilePath(TopDirPath,
                                   AbstractOneRefBowtie2IndexFilePath):
    """ Bowtie2 index file for an input FASTA file named after one
    reference """


class OneRefBowtie2IndexTempFilePath(StepDirPath,
                                     AbstractOneRefBowtie2IndexFilePath):
    """ Bowtie2 index file for a temporary FASTA file named after one
    reference """


class OneRefBowtie2IndexOutFilePath(ModuleDirPath,
                                    AbstractOneRefBowtie2IndexFilePath):
    """ Bowtie2 index file for an output FASTA file named after one
    reference """


class AbstractOneRefBowtie2IndexPrefix(BasePath, OneRefSeg):
    """ Abstract Bowtie2 index prefix for a FASTA file named after one
    reference """


class OneRefBowtie2IndexInPrefix(TopDirPath,
                                 AbstractOneRefBowtie2IndexPrefix):
    """ Bowtie2 index file for an input FASTA file named after one
    reference """


class OneRefBowtie2IndexTempPrefix(StepDirPath,
                                   AbstractOneRefBowtie2IndexPrefix):
    """ Bowtie2 index file for a temporary FASTA file named after one
    reference """


class OneRefBowtie2IndexOutPrefix(ModuleDirPath,
                                  AbstractOneRefBowtie2IndexPrefix):
    """ Bowtie2 index file for an output FASTA file named after one
    reference """


# Sequencing reads (FASTQ) file paths

class AbstractSampleReadsFilePath(BasePath, SampleReadsFileSeg):
    """ Abstract FASTQ file named after a sample """


class SampleReadsInFilePath(TopDirPath, AbstractSampleReadsFilePath):
    """ Input FASTQ file named after a sample """


class SampleReadsTempFilePath(StepDirPath, AbstractSampleReadsFilePath):
    """ Temporary FASTQ file named after a sample """


class SampleReadsOutFilePath(ModuleDirPath, AbstractSampleReadsFilePath):
    """ Output FASTQ file named after a sample """


class AbstractSampleReads1FilePath(BasePath, SampleReads1FileSeg):
    """ Abstract FASTQ mate 1 file named after a sample """


class SampleReads1InFilePath(TopDirPath, AbstractSampleReads1FilePath):
    """ Input FASTQ mate 1 file named after a sample """


class SampleReads1TempFilePath(StepDirPath, AbstractSampleReads1FilePath):
    """ Temporary FASTQ mate 1 file named after a sample """


class SampleReads1OutFilePath(ModuleDirPath, AbstractSampleReads1FilePath):
    """ Output FASTQ mate 1 file named after a sample """


class AbstractSampleReads2FilePath(BasePath, SampleReads2FileSeg):
    """ Abstract FASTQ mate 2 file named after a sample """


class SampleReads2InFilePath(TopDirPath, AbstractSampleReads2FilePath):
    """ Input FASTQ mate 2 file named after a sample """


class SampleReads2TempFilePath(StepDirPath, AbstractSampleReads2FilePath):
    """ Temporary FASTQ mate 2 file named after a sample """


class SampleReads2OutFilePath(ModuleDirPath, AbstractSampleReads2FilePath):
    """ Output FASTQ mate 2 file named after a sample """


class AbstractOneRefReadsFilePath(BasePath, OneRefReadsFileSeg):
    """ Abstract FASTQ file named after one reference """


class OneRefReadsInFilePath(SampleInDirPath, AbstractOneRefReadsFilePath):
    """ Input FASTQ file named after one reference """


class OneRefReadsTempFilePath(SampleStepDirPath, AbstractOneRefReadsFilePath):
    """ Temporary FASTQ file named after one reference """


class OneRefReadsOutFilePath(SampleOutDirPath, AbstractOneRefReadsFilePath):
    """ Output FASTQ file named after one reference """


class AbstractOneRefReads1FilePath(BasePath, OneRefReads1FileSeg):
    """ Abstract FASTQ mate 1 file named after one reference """


class OneRefReads1InFilePath(SampleInDirPath, AbstractOneRefReads1FilePath):
    """ Input FASTQ mate 1 file named after one reference """


class OneRefReads1TempFilePath(SampleStepDirPath, AbstractOneRefReads1FilePath):
    """ Temporary FASTQ mate 1 file named after one reference """


class OneRefReads1OutFilePath(SampleOutDirPath, AbstractOneRefReads1FilePath):
    """ Output FASTQ mate 1 file named after one reference """


class AbstractOneRefReads2FilePath(BasePath, OneRefReads2FileSeg):
    """ Abstract FASTQ mate 2 file named after one reference """


class OneRefReads2InFilePath(SampleInDirPath, AbstractOneRefReads2FilePath):
    """ Input FASTQ mate 2 file named after one reference """


class OneRefReads2TempFilePath(SampleStepDirPath, AbstractOneRefReads2FilePath):
    """ Temporary FASTQ mate 2 file named after one reference """


class OneRefReads2OutFilePath(SampleOutDirPath, AbstractOneRefReads2FilePath):
    """ Output FASTQ mate 2 file named after one reference """


# FASTQ quality control (FASTQC) paths

class FastqcOutDirPath(SampleOutDirPath, FastqcSeg):
    """ Output directory of FASTQC reports """


# Alignment map (SAM/BAM/CRAM) and index (BAM.BAI) file paths

class AbstractRefsetAlignmentFilePath(BasePath, RefsetAlignmentFileSeg):
    """ Abstract alignment map file named after a set of references """


class RefsetAlignmentInFilePath(SampleInDirPath,
                                AbstractRefsetAlignmentFilePath):
    """ Input alignment map file named after a set of references """


class RefsetAlignmentTempFilePath(SampleStepDirPath,
                                  AbstractRefsetAlignmentFilePath):
    """ Temporary alignment map file named after a set of references """


class RefsetAlignmentOutFilePath(SampleOutDirPath,
                                 AbstractRefsetAlignmentFilePath):
    """ Output alignment map file named after a set of references """


class AbstractRefsetAlignmentIndexFilePath(BasePath,
                                           RefsetAlignmentIndexFileSeg):
    """ Abstract alignment map index file named after a set of
    references """


class RefsetAlignmentIndexInFilePath(SampleInDirPath,
                                     AbstractRefsetAlignmentIndexFilePath):
    """ Input alignment map index file named after a set of
    references """


class RefsetAlignmentIndexTempFilePath(SampleStepDirPath,
                                       AbstractRefsetAlignmentIndexFilePath):
    """ Temporary alignment map index file named after a set of
    references """


class RefsetAlignmentIndexOutFilePath(SampleOutDirPath,
                                      AbstractRefsetAlignmentIndexFilePath):
    """ Output alignment map index file named after a set of
    references """


class AbstractOneRefAlignmentFilePath(BasePath, OneRefAlignmentFileSeg):
    """ Abstract alignment map file named after one reference """


class OneRefAlignmentInFilePath(SampleInDirPath,
                                AbstractOneRefAlignmentFilePath):
    """ Input alignment map file named after one reference """


class OneRefAlignmentTempFilePath(SampleStepDirPath,
                                  AbstractOneRefAlignmentFilePath):
    """ Temporary alignment map file named after one reference """


class OneRefAlignmentOutFilePath(SampleOutDirPath,
                                 AbstractOneRefAlignmentFilePath):
    """ Output alignment map file named after one reference """


class AbstractOneRefAlignmentIndexFilePath(BasePath,
                                           OneRefAlignmentIndexFileSeg):
    """ Abstract alignment map index file named after one reference """


class OneRefAlignmentIndexInFilePath(SampleInDirPath,
                                     AbstractOneRefAlignmentIndexFilePath):
    """ Input alignment map index file named after one reference """


class OneRefAlignmentIndexTempFilePath(SampleStepDirPath,
                                       AbstractOneRefAlignmentIndexFilePath):
    """ Temporary alignment map index file named after one reference """


class OneRefAlignmentIndexOutFilePath(SampleOutDirPath,
                                      AbstractOneRefAlignmentIndexFilePath):
    """ Output alignment map index file named after one reference """


class AbstractSectionAlignmentFilePath(BasePath, SectionAlignmentFileSeg):
    """ Abstract alignment map file named after a section """
    ref: str


class SectionAlignmentInFilePath(RefInDirPath,
                                 AbstractSectionAlignmentFilePath):
    """ Input alignment map file named after a section """


class SectionAlignmentTempFilePath(RefStepDirPath,
                                   AbstractSectionAlignmentFilePath):
    """ Temporary alignment map file named after a section """


class SectionAlignmentOutFilePath(RefOutDirPath,
                                  AbstractSectionAlignmentFilePath):
    """ Output alignment map file named after a section """


# Vectoring file paths

class MutVectorBatchFilePath(SectionOutDirPath, MutVectorBatchFileSeg):
    """ Output file of a batch of mutation vectors """


class MutVectorReportFilePath(SectionOutDirPath, MutVectorReportFileSeg):
    """ Output vectorization report file """


# Path collection and conversion functions #############################

def is_concrete_path_class(member: Any):
    """ Return whether ```member``` is a class of concrete path. """
    # To be a class of concrete path (i.e. instantiable, with at least
    # the top-level field and all essential path methods implemented),
    # a member of the module must 1) be a class and 2) be TopDirPath or
    # a subclass of TopDirPath (issubclass returns True for both), which
    # provides the top-level segment and all required methods of paths.
    return isclass(member) and issubclass(member, TopDirPath)


@cache
def _generate_name_to_path_class() -> dict[str, type[BasePath]]:
    """ Return a ```dict``` of every valid, instantiable class of path
    in this module, with class names as keys. """
    return dict(getmembers(sys.modules[__name__], is_concrete_path_class))


@cache
def _generate_fields_exts_to_path_class():
    """ Return a ```dict``` of every valid, instantiable class of path.
    Each key is a ```tuple``` of field names in alphabetical order.
    For directories, the field names uniquely determine the path class.
    For files, the extension also determines the path class; thus, each
    key of field names maps to another ```dict``` keyed by extensions.
    One entry each for a directory and set of files are illustrated:
    >>> _ = {
    ...     # Fields 'top' and 'module' map uniquely to ModuleDirPath.
    ...     ('module', 'top'): ModuleDirPath,
    ...     # Fields 'top', 'module', 'sample', 'ref', 'end5', 'end3',
    ...     # and 'ext' map to four file extensions that map to a total
    ...     # of two classes of files (```SectionAlignmentOutFilePath```
    ...     # and ```MutVectorReportFilePath```).
    ...     ('end3', 'end5', 'ext', 'module', 'ref', 'sample', 'top'): {
    ...         '.json': MutVectorReportFilePath,
    ...         '.bam': SectionAlignmentOutFilePath,
    ...         '.cram': SectionAlignmentOutFilePath,
    ...         '.sam': SectionAlignmentOutFilePath,
    ...     }
    ... }
    """
    # Dictionary that maps fields and extensions to path classes
    fe_to_cls = dict()
    # Iterate over every valid class of path.
    for cls in _generate_name_to_path_class().values():
        # Get the field names of the class in alphabetical order.
        fields = tuple(sorted(cls.keys()))
        if EXT_KEY in fields:
            # The class has a field for file extensions.
            if fields not in fe_to_cls:
                # If the dictionary of extensions for this set of fields
                # does not yet exist, initialize an empty one.
                fe_to_cls[fields]: dict[str, type[BasePath]] = dict()
            if not isinstance(fe_to_cls[fields], dict):
                raise TypeError(f"Expected type 'dict' at {fields}, but got "
                                f"'{type(fe_to_cls[fields]).__name__}'")
            # For each file extension of the class, add the class to the
            # dictionary using the extension as the key.
            for ext in getattr(cls, EXTS_KEY):
                if ext in fe_to_cls[fields]:
                    raise ValueError(f"{fe_to_cls[fields]} and {cls} have "
                                     f"identical fields {fields} and file "
                                     f"extension '{ext}'")
                fe_to_cls[fields][ext] = cls
        else:
            # The class does not have a field for file extensions.
            if fields in fe_to_cls:
                raise TypeError(f"Cannot insert {cls.__name__} at {fields}: "
                                f"{fe_to_cls[fields]} is already there")
            # Add the class to the dictionary, keyed only by the fields.
            fe_to_cls[fields] = cls
    return fe_to_cls


def get_path_class_by_name(name: str):
    """ Return the path class with the given name. """
    try:
        return _generate_name_to_path_class()[name]
    except KeyError:
        raise ValueError(f"No path class named '{name}'")


def get_path_class_by_fields_ext(fields: Iterable[str],
                                 ext: Iterable[str] | str | None = None):
    """ Return the path class with the given fields and, if applicable,
    file extension. If no file extension is given, the class is assumed
    to be a directory. """
    # Get the fields in standard form (tuple, alphabetical order).
    std_fields = tuple(sorted(fields))
    if ext is None:
        # Assume the class is a directory path.
        try:
            return _generate_fields_exts_to_path_class()[std_fields]
        except KeyError:
            raise ValueError(f"No class of directory with fields {std_fields}")
    # The class is not a directory: assume it is a file path.
    if isinstance(ext, str):
        # Search for the file path given a single file extension.
        try:
            return _generate_fields_exts_to_path_class()[std_fields][ext]
        except KeyError:
            raise ValueError(f"No class of file with fields {std_fields} and "
                             f"extension '{ext}'")
    # Search for the file path given multiple file extensions.
    # First, find the type of path corresponding to each extension.
    # Keep at most one of each type of path by storing them in a set
    # before converting the set to a list.
    classes = list({get_path_class_by_fields_ext(std_fields, e) for e in ext})
    # Verify that exactly one unique type of path corresponds to all
    # the given file extensions.
    if not classes:
        raise ValueError(f"No path class with fields {std_fields} and "
                         f"extensions {ext}: {classes}")
    if len(classes) > 1:
        raise ValueError(f"Multiple path classes with fields {std_fields} and "
                         f"extensions {ext}: {classes}")
    # Return the type of path corresponding to all the file extensions.
    return classes[0]


def clean_fields(fields: dict[str, Any]):
    """ Remove each field whose value is ```None```. """
    return {name: value for name, value in fields.items() if value is not None}


def get_path_class_by_fields(**fields):
    """ Return the path class that corresponds to the given fields. Like
    ```get_path_class_by_fields_ext```, but here all fields are given as
    keyword arguments rather than an iterable of fields and a separate
    argument for the optional file extension. """
    return get_path_class_by_fields_ext(clean_fields(fields),
                                        fields.get(EXT_KEY))


def create(**fields):
    """ Create a new path instance with the given fields. """
    return get_path_class_by_fields(**fields)(**clean_fields(fields))


# Ensure that there are no two paths with identical fields.
_generate_fields_exts_to_path_class()


from ..draw import manipulator, util, plotter
import pandas as pd
import numpy as np
from ..util.dump import sort_dict, flatten_json
import plotly.graph_objects as go
from custom_inherit import doc_inherit
from ..util.docstring import style_child_takes_over_parent
import os
from .util import save_plot, extract_args
import inspect 

class Study(object):
    """A class to store information about a study, i.e a set of samples that are relevant to be studied together.

    Attributes:
        name (str, optional): Short description (<~20 char) of your study. Defaults to None.
        samples (List[str], optional): Names of your study's samples. Defaults to None.
        
    Example:
        >>> study = Study('example',['A1', 'B2', 'B3'])
    """

    attr_list = ['name','samples']

    def __init__(self, data=None, min_cov=0, filter_by='sample') -> None:
        """Creates a Study object.

        Args:
            data (dict or list[dict] or pandas.DataFrame, optional): Data to use. Can be a dictionary or list of dictionaries containing DREEM-output jsons, or directly a pandas dataframe. Defaults to None.
            min_cov (int, optional): Minimum number of base coverage for a row to be filtered-in. Defaults to 0.
            filter_by (str, optional): Filter rows by sample or study. When filtered by study, if a row passes the filter, rows with the same 'reference', 'section' and 'cluster' fields for all other samples have a sufficient base coverage. Defaults to 'sample'.            

        Example:
            >>> study = Study(data = {'sample':'mysample',{'reference1': {'section1': {'cluster1': {'sub_N': [100], 'cov': [1000]}}}}},
                              min_cov=1000, 
                              filter_by='sample')
            >>> study = Study(data = (
                                        {'sample':'mysample',{'reference1': {'section1': {'cluster1': {'sub_N': [100], 'cov': [1000]}}}}},
                                        {'sample':'mysample2',{'reference1': {'section1': {'cluster1': {'sub_N': [99], 'cov': [1000]}}}}},
                                        ),
                              min_cov=1000, 
                              filter_by='sample')      
        """
        if data is not None:
            
            df = pd.DataFrame()              
            
            # If data is a list of json, concatenate them into a single dataframe
            if type(data) is not pd.DataFrame:
                print('Turning data into a dataframe...')
                                
                # if data isn't iterable, make it a list
                if not hasattr(data, '__iter__') or isinstance(data, dict):
                    data = [data]
                
                for sample in data:
                    print(sample['sample'], end='... ')
                    df = pd.concat([df, pd.DataFrame(flatten_json(sort_dict(sample)))], axis=0)
                
                print('Done.')
            
            # Use the dataframe (loaded or created from json)
            print('Setting dataframe...')
            self.set_df(df, min_cov=min_cov, filter_by=filter_by)
            print('Done.')
            
        else:
            self.df = None

    
    def set_df(self, df, min_cov=0, filter_by='sample'):
        
        self.df = df.reset_index(drop=True)
                
        self.df = self.df[self.df['min_cov'] >= min_cov]
        
        if filter_by == 'study':
            self.filter_by_study()
        
        for attr in ['sample','reference']:
            self.df[attr] = self.df[attr].astype(str)
        
        for attr in ['section','cluster']:
            if attr not in self.df.columns:
                self.df[attr] = 0
        
        self.df['deltaG'] = self.df['deltaG'].apply(lambda x: 0.0 if x == 'void' else float(x))
    
    
    def filter_by_study(self):
        df = self.df.groupby(['reference', 'section', 'cluster']).filter(lambda x: len(self.df['sample'].unique()) == len(x['sample'].unique()))
        self.df = df

    def get_samples(self):
        return self.df['sample'].unique()

    def get_references(self, sample:str):
        return self.df[self.df['sample'] == sample]['reference'].unique()

    def get_sections(self, sample:str, reference:str):
        return self.df[(self.df['sample'] == sample) & (self.df['reference'] == reference)]['section'].unique()

    def get_clusters(self, sample:str, reference:str, section:str):
        return self.df[(self.df['sample'] == sample) & (self.df['reference'] == reference)& (self.df['section'] == section)]['cluster'].unique()
    
    def wrap_to_plotter(self, func, loc, kwargs):

        kwargs = {
                **{k:v for k,v in loc.items() if not k in ['self', 'args', 'kwargs']},
                **kwargs
            }

        """Wrapper for the plot functions."""
        return func(
            manipulator.get_df(self.df, 
                                **{k:v for k,v in kwargs.items() if k in list(self.df.columns)+ extract_args(manipulator.get_df)}), 
                                **{k:v for k,v in kwargs.items() if k in extract_args(func)})

    
    def default_arguments_per_base(self):
        """Default arguments for the plot functions.
        
        Args:
            base_index (list[int], list[str], optional): Filter per-base attributes (sub_rate, sequence, etc) by base index. Can be a unique sequence in the row's sequence, a list of indexes or a single index. Defaults to None.
            base_type (list, str, optional): Filter per-base attributes (sub_rate, sequence, etc) by base type. Defaults to ``['A','C','G','T']``.
            base_pairing (bool, optional): Filter per-base attributes (sub_rate, sequence, etc) by expected base pairing. True will keep only base pairs, False will keep only non-base pairs. Defaults to None.
            **kwargs: Additional arguments to pass to filter rows by. Ex: ``flank='flank_1'`` will keep only rows with ``flank==flank_1``. 

        Returns:
            dict: {``'fig'``: a plotly figure, ``'data'``: a pandas dataframe}
            
        """
    

    @doc_inherit(default_arguments_per_base, style=style_child_takes_over_parent)
    def default_arguments_single_row(self):
        """Default arguments for the mutiple rows plot functions.
        
        Args:
            sample (str, optional): Selects this sample. Defaults to None.
            reference (str, optional): Selects this reference. Defaults to None.
            section (str, optional): Selects this section. Defaults to ``full``.
            cluster (str, optional): Selects this cluster. Defaults to ``pop_avg``.
            
        """
 
    @doc_inherit(default_arguments_per_base, style=style_child_takes_over_parent)
    def default_arguments_multi_rows(self):
        """Default arguments for the single row plot functions.
        
        Args:
            sample (list, str, optional): Filter rows by sample (a list of samples or just a sample). Defaults to None.
            reference (list, str, optional): Filter rows by reference (a list of references or just a reference). Defaults to None.
            section (list, str, optional): Filter rows by section (a list of sections or just a section). Defaults to None.
            cluster (list, str, optional): Filter rows by cluster (a list of clusters or just a cluster). Defaults to None.
            
        """
        
    @doc_inherit(default_arguments_per_base, style=style_child_takes_over_parent)
    @doc_inherit(default_arguments_multi_rows, style=style_child_takes_over_parent)
    def get_df(self, **kwargs):
        """Filter the dataframe by the given arguments."""
        return manipulator.get_df(self.df, **kwargs)

    
    ############################################################################################################
    # Plot functions                                                                                           #
    ############################################################################################################
    
    @save_plot
    @doc_inherit(save_plot, style=style_child_takes_over_parent)
    @doc_inherit(default_arguments_single_row, style=style_child_takes_over_parent)
    def mutation_fraction(self, sample, reference, section=None, cluster=None,  **kwargs)->dict:
        """Plot the mutation rates as histograms.

        Args:
            show_ci(bool, optional): Show confidence intervals. Defaults to True.
            
        """
        return self.wrap_to_plotter(
            plotter.mutation_fraction,
            locals(),
            kwargs
        )
        
    
    @save_plot
    @doc_inherit(save_plot, style=style_child_takes_over_parent)
    @doc_inherit(default_arguments_single_row, style=style_child_takes_over_parent)
    def mutation_fraction_identity(self, sample, reference, section=None, cluster=None,  **kwargs)->dict:
        """Plot the mutation rates as histograms.

        Args:
            show_ci(bool, optional): Show confidence intervals. Defaults to True.
            
        """
        return self.wrap_to_plotter(
            plotter.mutation_fraction_identity,
            locals(),
            kwargs
        )

    @save_plot
    @doc_inherit(save_plot, style=style_child_takes_over_parent)
    @doc_inherit(default_arguments_multi_rows, style=style_child_takes_over_parent)
    def deltaG_vs_sub_rate(self, **kwargs)->dict:
        """Plot the Mutation fraction of each paired-expected base of the ROI for each reference of a sample, w.r.t the deltaG estimation.

        Args:
            models (List[str], optional): Models to fit on the data using scipy.optimize.curve_fit. Under the form ``'lambda x, a, b: a*x+b'`` where ``x`` is the variable. Defaults to [].

        """
        return self.wrap_to_plotter(
            plotter.deltaG_vs_sub_rate,
            locals(),
            kwargs
        )

    @save_plot
    @doc_inherit(save_plot, style=style_child_takes_over_parent)
    @doc_inherit(default_arguments_multi_rows, style=style_child_takes_over_parent)
    def experimental_variable_across_samples(self, experimental_variable, reference, section, **kwargs)->dict:
        """Plot a given experimental variable vs Mutation fraction across samples for a given reference and section.

        Args:
            experimental_variable (str): Name of the experimental variable to plot.
            models (List[str], optional): Models to fit on the data using scipy.optimize.curve_fit. Under the form ``'lambda x, a, b: a*x+b'`` where ``x`` is the variable. Defaults to [].
            
        """
        index_selected = True
        return self.wrap_to_plotter(
            plotter.experimental_variable_across_samples,
            locals(),
            kwargs
        )

    # @save_plot
    # @doc_inherit(save_plot, style=style_child_takes_over_parent)
    # @doc_inherit(default_arguments_multi_rows, style=style_child_takes_over_parent)
    # def auc(self, **kwargs)->dict:
    #     """Plot the AUC for each mutation profile of the selected data. 

    #     """
    #     unique_id = True
    #     return self.wrap_to_plotter(
    #         plotter.auc,
    #         locals(),
    #         kwargs
    #     )

    @save_plot
    @doc_inherit(save_plot, style=style_child_takes_over_parent)
    @doc_inherit(default_arguments_multi_rows, style=style_child_takes_over_parent)
    def mutations_in_barcodes(self, sample, section='barcode', **kwargs)->dict:
        """Plot the number of mutations in the barcode per read of a sample as an histogram.

        """
        return self.wrap_to_plotter(
            plotter.mutations_in_barcodes,
            locals(),
            kwargs
        )
                    
    @save_plot
    @doc_inherit(save_plot, style=style_child_takes_over_parent)  
    @doc_inherit(default_arguments_multi_rows, style=style_child_takes_over_parent)
    def num_aligned_reads_per_reference_frequency_distribution(self, sample, section=None, **kwargs)->dict:
        """Plot the number of aligned reads per reference as a frequency distribution. x axis is the number of aligned reads per reference, y axis is the count of reference that have this number of aligned reads.

        """
        
        return self.wrap_to_plotter(
            plotter.num_aligned_reads_per_reference_frequency_distribution,
            locals(),
            kwargs
        )        

    @save_plot
    @doc_inherit(save_plot, style=style_child_takes_over_parent)
    def mutation_fraction_delta(self, **kwargs)->dict:
        """Plot the Mutation fraction difference between two mutation profiles.
        
        Args:
            sample1: sample of the first mutation profile.
            sample2: sample of the second mutation profile.
            reference1: reference of the first mutation profile.
            reference2: reference of the second mutation profile.
            section1: section of the first mutation profile.
            section2: section of the second mutation profile.
            cluster1: cluster of the first mutation profile.
            cluster2: cluster of the second mutation profile.
            base_index1: base index of the first mutation profile.
            base_index2: base index of the second mutation profile.
            base_type1: base type of the first mutation profile.
            base_type2: base type of the second mutation profile.
            base_pairing1: base pairing of the first mutation profile.
            base_pairing2: base pairing of the second mutation profile. 
        
        Returns:
            dict: {'fig': a plotly figure, 'data': a pandas dataframe}
        
        """
        
        df1 = manipulator.get_df(self.df, **{k[:-1]:v for k,v in kwargs.items() if k.endswith('1') and k[:-1] in list(self.df.columns)+ list(manipulator.get_df.__code__.co_varnames)})
        assert len(df1)>0, 'No rows found for the first mutation profile.'
        assert len(df1)==1, 'More than one row found for the first mutation profile.'
        df2 = manipulator.get_df(self.df, **{k[:-1]:v for k,v in kwargs.items() if k.endswith('2') and k[:-1] in list(self.df.columns)+ list(manipulator.get_df.__code__.co_varnames)})
        assert len(df2)>0, 'No rows found for the second mutation profile.'
        assert len(df2)==1, 'Only one row should be selected for the second mutation profile.'
        return plotter.mutation_fraction_delta(pd.concat([df1, df2]).reset_index(drop=True), **{k:v for k,v in kwargs.items() if k in plotter.mutation_fraction_delta.__code__.co_varnames})

    @save_plot
    @doc_inherit(save_plot, style=style_child_takes_over_parent)
    @doc_inherit(default_arguments_multi_rows, style=style_child_takes_over_parent)
    def mutations_per_read_per_sample(self, sample, section=None, **kwargs)->dict:
        """Plot the number of mutations per read per sample as an histogram.

        """
        return self.wrap_to_plotter(
            plotter.mutations_per_read_per_sample,
            locals(),
            kwargs
        )

    @save_plot
    @doc_inherit(save_plot, style=style_child_takes_over_parent)
    @doc_inherit(default_arguments_single_row, style=style_child_takes_over_parent)
    def base_coverage(self, sample, reference, section=None, cluster=None, **kwargs):
        """Plot the base coverage of a single row of your dataframe.

        """
        return self.wrap_to_plotter(
            plotter.base_coverage,
            locals(),
            kwargs
        )

    
    @save_plot
    @doc_inherit(save_plot, style=style_child_takes_over_parent)
    @doc_inherit(default_arguments_single_row, style=style_child_takes_over_parent)
    def mutation_per_read_per_reference(self, sample, reference, section=None, cluster=None, **kwargs)->dict:
        """Plot the number of mutations per read per reference as an histogram.

        """
        return self.wrap_to_plotter(
            plotter.mutation_per_read_per_reference,
            locals(),
            kwargs
        )

    @save_plot
    @doc_inherit(save_plot, style=style_child_takes_over_parent)
    @doc_inherit(default_arguments_multi_rows, style=style_child_takes_over_parent)
    def compare_mutation_profiles(self, max_plots = 100, max_axis = None, **kwargs):
        """Plot the mutation fraction of multiple mutation profiles.

        Args:
            max_plots: maximum number of plots to show.
            max_axis: maximum value of the x and y axis. If None, the maximum value of the data will be used if above 0.15, otherwise 0.15.
        """
        kwargs['unique_id'] = True
        
        return self.wrap_to_plotter(
            plotter.compare_mutation_profiles,
            locals(),
            kwargs
        )



from hashlib import md5
import logging
import os
from typing import Any

import numpy as np
import pandas as pd


def get_files(folder: str, ext: str):
    paths = []
    for reference in os.listdir(folder):
        if os.path.isdir(os.path.join(folder, reference)):
            for section in os.listdir(os.path.join(folder, reference)):
                if not section.endswith(ext):
                    continue
                paths.append(os.path.join(folder, reference, section))
    else:
        if reference.endswith(ext):
            paths.append(os.path.join(folder, reference))
    return paths


def fastq_to_df(fastq_file):    
    df, data = pd.DataFrame(), pd.read_csv(fastq_file, sep='\t', header=None)
    df['reference'] = data.iloc[np.arange(0,len(data),4)].reset_index(drop=True)
    df['sequence'] = data.iloc[np.arange(1,len(data),4)].reset_index(drop=True)
    df['quality'] = data.iloc[np.arange(3,len(data),4)].reset_index(drop=True)
    return df


def sam_to_df(path):
    skiprows = 0
    with open(path) as f:
        while f.readline().startswith('@'):
            skiprows += 1
    df = pd.read_csv(path, sep='\t', skiprows=skiprows, header=None)
    df = df[df.columns[:11]]
    df.columns = ['QNAME', 'FLAG', 'RNAME', 'POS', 'MAPQ', 'CIGAR', 'RNEXT', 'PNEXT','TLEN', 'SEQ', 'QUAL']
    return df


def sort_fastq_pairs(fq1s, fq2s):
    if type(fq1s) is str:
        fq1s = [fq1s]
    if type(fq2s) is str:
        fq2s = [fq2s]
    assert len(fq1s) >= len(fq2s), 'More fq2s than fq1s'
    for f2 in fq2s:
        if f2.replace('_R2', '_R1') not in fq1s:
            raise ValueError(f'No matching pair for {f2}')
    fq1s = [f2.replace('_R2', '_R1') for f2 in fq2s] + [f for f in fq1s if f.replace('_R1','_R2') not in fq2s]
    fq2s += [None for f in fq1s if f.replace('_R1','_R2') not in fq2s]
    samples = [f.split('/')[-1].split('.')[0].split('_')[:-1] for f in fq1s]
    samples = ['_'.join(s) for s in samples]
    return fq1s, fq2s, samples


def query_muts(muts: np.ndarray, bits: int, sum_up = True, axis=0, set_type = 'superset'):
    """
    Count the number of times a query mutation occurs in each column
    or one column of a set of mutation mut_vectors.
    The counting operation comprises three steps:
    1. bitwise AND to confirm at least one "1" bit is shared, e.g.
       bits: 11110000 & mut_vectors: 00100000 -> 00100000 (True)
       bits: 11110000 & mut_vectors: 00100010 -> 00100000 (True)
       bits: 11110000 & mut_vectors: 00000000 -> 00000000 (False)
    2. bitwise OR to confirm no "1" bit in mut_vectors is not in bits, e.g.
       bits: 11110000 | mut_vectors: 00100000 -> 11110000 =? 11110000 (True)
       bits: 11110000 | mut_vectors: 00100010 -> 11110010 =? 11110000 (False)
       bits: 11110000 | mut_vectors: 00000000 -> 11110000 =? 11110000 (True)
    3. logical AND to confirm that both test pass, e.g.
       bits: 11110000, mut_vectors: 00100000 -> True  AND True  (True)
       bits: 11110000, mut_vectors: 00100010 -> True  AND False (False)
       bits: 11110000, mut_vectors: 00000000 -> False AND True  (False)

    Arguments
    mut_vectors: NDArray of a set of mutation mut_vectors (2-dimensional)
          or one column in a set of mutation mut_vectors (1-dimensional).
          Data type must be uint8.
    bits: One-byte int in the range [0, 256) representing the mutation
          to be queried. The bits in the int encode the mutation as
          defined above, e.g.
          - 00000010 (int 2) is a deletion
          - 11010001 (int 209) is either substitution to A, G, or T
                               or a match to C
    
    Returns
    if sum_up: 
        int of the number of times the query mutation occurs in mut_vectors
        count: If mut_vectors is 1-dimensional, int of the number of times the
            query mutation occurs in mut_vectors.
            If mut_vectors is 2-dimensional, NDArray with one int for each
            column in mut_vectors.
    if not sum_up:
        bool NDArray with one bool for each column in mut_vectors.
        True if the query mutation occurs in the column, False otherwise.
    """
    if not muts.dtype == np.uint8:
        raise TypeError('mut_vectors must be of type uint8 and not {}'.format(muts.dtype))
    #print(np.logical_and(mut_vectors & bits, (mut_vectors | bits) == bits).sum(axis=0))
    assert isinstance(bits, int) and 0 <= bits < 256

        
    if set_type == 'subset':
        logic_fun = lambda m, b: np.logical_and(m & b, (m | b) == b)
    if set_type == 'superset':
        logic_fun = lambda m, b: np.array(m & b, dtype=bool)
    
    if sum_up:
        return logic_fun(muts, bits).sum(axis=axis)
    else:
        return logic_fun(muts, bits)


def digest_file(file_path: Any) -> str:
    """
    Compute the checksum of a file.

    Parameters
    ----------
    file_path: Any (path-like)
        Path of the file on which to compute the checksum. Can be
        any type that the open() function recognizes as a path.

    Returns
    -------
    str
        Checksum of the file (in hexadecimal)
    """
    with open(file_path, "rb") as f:
        digest = md5(f.read()).hexdigest()
    return digest


def astuple(item: Any) -> tuple:
    """ Convert an item to a ```tuple```. """
    if isinstance(item, tuple):
        # If item is already a tuple, return it unmodified.
        return item
    try:
        # If item is iterable, then convert it to a tuple.
        return tuple(item)
    except TypeError:
        # If item is not iterable, then put it in a length-1 tuple.
        return item,


def aslist(item: Any) -> list:
    """ Convert an item to a ```list```. """
    if isinstance(item, list):
        # If item is already a list, return a copy to avoid modifying
        # the original list.
        return item.copy()
    try:
        # If item is iterable, then convert it to a list.
        return list(item)
    except TypeError:
        # If item is not iterable, then put it in a length-1 list.
        return [item]



from logging import getLogger
from pathlib import Path
from typing import Iterable

from ..util import path


logger = getLogger(__name__)

# FASTA format
FASTA_NAMESYM = b">"

# Byte encodings for nucleic acid alphabets
BASES = b"ACGT"
COMPS = b"TGCA"
RBASE = b"ACGU"
RCOMP = b"UGCA"
BASEN = b"N"
BASES_SET = set(BASES)
BASEN_SET = set(BASES + BASEN)

# Integer encodings for nucleic acid alphabets
A_INT = BASES[0]
C_INT = BASES[1]
G_INT = BASES[2]
T_INT = BASES[3]
N_INT = BASEN[0]

# Integer encodings for mutation vectors
BLANK = b"\x00"[0]  # 00000000 (000): no coverage at this position
MATCH = b"\x01"[0]  # 00000001 (001): match with reference
DELET = b"\x02"[0]  # 00000010 (002): deletion from reference
INS_5 = b"\x04"[0]  # 00000100 (004): insertion 5' of base in reference
INS_3 = b"\x08"[0]  # 00001000 (008): insertion 3' of base in reference
SUB_A = b"\x10"[0]  # 00010000 (016): substitution to A
SUB_C = b"\x20"[0]  # 00100000 (032): substitution to C
SUB_G = b"\x40"[0]  # 01000000 (064): substitution to G
SUB_T = b"\x80"[0]  # 10000000 (128): substitution to T
SUB_N = SUB_A | SUB_C | SUB_G | SUB_T
ANY_N = SUB_N | MATCH
INDEL = DELET | INS_5 | INS_3
EVERY = ANY_N | INDEL


def get_diffs(seq1, seq2):
    if len(seq1) != len(seq2):
        raise ValueError("Sequences were different lengths: "
                         f"'{seq1}' and '{seq2}'")
    diffs = [i for i, (x1, x2) in enumerate(zip(seq1, seq2)) if x1 != x2]
    return diffs


class Seq(bytes):
    __slots__ = []

    alph = b""
    comp = b""
    alphaset = set(alph)
    comptrans = alph.maketrans(alph, comp)

    def __init__(self, seq: bytes):
        self.validate_seq(seq)
        super().__init__()
    
    @classmethod
    def validate_seq(cls, seq):
        if not seq:
            raise ValueError("seq is empty")
        if set(seq) - cls.alphaset:
            raise ValueError(f"Invalid characters in seq: '{seq.decode()}'")

    @property
    def rc(self):
        return self.__class__(self[::-1].translate(self.comptrans))

    def __getitem__(self, item):
        return self.__class__(super().__getitem__(item))

    def __str__(self):
        return self.decode()


class DNA(Seq):
    alph = BASES
    comp = COMPS
    alphaset = set(alph)
    comptrans = alph.maketrans(alph, comp)

    def tr(self):
        """ Transcribe DNA to RNA. """
        return RNA(self.replace(b"T", b"U"))


class RNA(Seq):
    alph = RBASE
    comp = RCOMP
    alphaset = set(alph)
    comptrans = alph.maketrans(alph, comp)

    def rt(self):
        """ Reverse transcribe RNA to DNA. """
        return DNA(self.replace(b"U", b"T"))


def parse_fasta(fasta: str | Path):
    """ Parse a FASTA file and iterate through the reference names and
    sequences. """
    if not fasta:
        raise TypeError("No FASTA file given")
    logger.info(f"Began parsing FASTA: {fasta}")
    # Get the name of the set of references.
    refset = path.RefsetSeqInFilePath.parse(fasta).refset
    has_ref_named_refset = False
    # Record the names of all the references.
    names = set()
    with open(fasta, "rb") as f:
        line = f.readline()
        while line:
            # Read the name from the current line.
            if not line.startswith(FASTA_NAMESYM):
                logger.error(f"Name line '{line.strip()}' in {fasta} does not "
                             f"start with name symbol '{FASTA_NAMESYM}'")
                continue
            # Get the name of the reference up to the first whitespace.
            name = line.split(maxsplit=1)[0][len(FASTA_NAMESYM):].decode()
            # Read the sequence of the reference up until the next
            # reference or the end of the file, whichever comes first.
            seqarray = bytearray()
            while (line := f.readline()) and not line.startswith(FASTA_NAMESYM):
                seqarray.extend(line.rstrip())
            # Confirm that the sequence is a valid DNA sequence.
            try:
                seq = DNA(seqarray)
            except Exception as error:
                logger.error(f"Failed to parse sequence in {fasta}: {error}")
                continue
            # Confirm that the name is not blank.
            if not name:
                logger.error(f"Blank name line '{line.strip()}' in {fasta}")
                continue
            # If there are two or more references with the same name,
            # then the sequence of only the first is used.
            if name in names:
                logger.warning(f"Duplicate reference '{name}' in {fasta}")
                continue
            # If any reference has the same name as the file, then the
            # file is not allowed to have any additional references
            # because, if it did, then the files of all references and
            # of only the self-named reference would have the same names
            # and thus be indistinguishable by their paths.
            if name == refset:
                has_ref_named_refset = True
                logger.debug(f"Reference '{name}' had same name as {fasta}")
            if has_ref_named_refset and names:
                raise ValueError(f"Because {fasta} had a reference with the "
                                 f"same name as the file ('{name}'), it was "
                                 f"not allowed to have any other references, "
                                 f"but it also had {', '.join(names)}")
            # Yield the validated name and sequence.
            names.add(name)
            logger.debug(f"Read reference '{name}' of length {len(seq)} "
                         f"from {fasta}")
            yield name, seq
    logger.info(f"Ended parsing {len(names)} references from FASTA: {fasta}")


def write_fasta(fasta: str | Path, refs: Iterable[tuple[str, DNA]]):
    """ Write an iterable of reference names and DNA sequences to a
    FASTA file. """
    if not fasta:
        raise TypeError("No FASTA file given")
    logger.info(f"Began writing FASTA file: {fasta}")
    # Get the name of the set of references.
    refset = path.RefsetSeqInFilePath.parse(fasta).refset
    has_ref_named_refset = False
    # Record the names of all the references.
    names = set()
    with open(fasta, "xb") as f:
        for name, seq in refs:
            # Confirm that the name is not blank.
            if not name:
                logger.error(f"Blank reference name")
                continue
            # If there are two or more references with the same name,
            # then the sequence of only the first is used.
            if name in names:
                logger.warning(f"Duplicate reference '{name}'")
                continue
            # If any reference has the same name as the file, then the
            # file is not allowed to have any additional references
            # because, if it did, then the files of all references and
            # of only the self-named reference would have the same names
            # and thus be indistinguishable by their paths.
            if name == refset:
                has_ref_named_refset = True
                logger.debug(f"Reference '{name}' had same name as {fasta}")
            if has_ref_named_refset and names:
                raise ValueError(f"Because {fasta} got a reference with the "
                                 f"same name as the file ('{name}'), it was "
                                 f"not allowed to get any other references, "
                                 f"but it also got {', '.join(names)}")
            try:
                f.write(b"".join([FASTA_NAMESYM, name.encode(), b"\n",
                                  seq, b"\n"]))
            except Exception as error:
                logger.error(
                    f"Error writing reference '{name}' to {fasta}: {error}")
            else:
                logger.debug(f"Wrote reference '{name}' of length {len(seq)}"
                             f"to {fasta}")
                names.add(name)
    logger.info(f"Wrote {len(names)} references to {fasta}")



import logging
import shlex
import subprocess
from typing import Any, List

from ..util.logs import log_process


logger = logging.getLogger(__name__)


# Commands for external applications
BOWTIE2_CMD = "bowtie2"
BOWTIE2_BUILD_CMD = "bowtie2-build"
CUTADAPT_CMD = "cutadapt"
FASTQC_CMD = "fastqc"
SAMTOOLS_CMD = "samtools"


# Command utility functions

def run_cmd(args: List[Any]):
    """ Run a command via subprocess.run(), with logging. """
    # Use shlex to place quotes around arguments containing whitespace.
    cmd = shlex.join(map(str, args))
    # Log the command with which the process was run.
    logger.debug(f"Shell $ {cmd}")
    # Run the process and capture the output.
    process = subprocess.run(cmd, check=True, shell=True, capture_output=True)
    # Log the output of the process.
    log_process(logger, process)
    return process
